import os # Cleaning console
from dataclasses import dataclass
from IPython.display import Markdown, display
from typing import Optional, Dict, Any
import matplotlib.pyplot as plt
from matplotlib.patches import Circle
from pint.facets.plain import PlainQuantity
import numpy as np
import pandas as pd
from pandas import DataFrame
import math
import warnings

from mento.rectangular import RectangularSection
from mento.material import Concrete, SteelBar, Concrete_ACI_318_19, Concrete_EN_1992_2004, Concrete_CIRSOC_201_25
from mento.rebar import Rebar
from mento.units import MPa, psi, mm, inch, kN, m, cm, kNm, dimensionless
from mento.results import Formatter, TablePrinter, DocumentBuilder, CUSTOM_COLORS
from mento.forces import Forces  

from mento.codes.EN_1992_2004_beam import _check_shear_EN_1992_2004, _design_shear_EN_1992_2004
from mento.codes.ACI_318_19_beam import _check_shear_ACI_318_19, _design_shear_ACI_318_19, _check_flexure_ACI_318_19, _design_flexure_ACI_318_19


@dataclass
class RectangularBeam(RectangularSection):
    

    def __init__(self, label: Optional[str], concrete: Concrete, steel_bar: SteelBar, 
                 width: PlainQuantity, height: PlainQuantity, settings: Optional[Dict[str, Any]] = None):   
        super().__init__(label, concrete, steel_bar, width, height, settings)
        if settings:
            self.settings.update(settings)  # Update with any provided settings
        
        self.layers_spacing = self.settings.get_setting('layers_spacing')
        
        # Centralized attribute initialization
        self._initialize_attributes()
        
        # self.shear_design_results: DataFrame = None

##########################################################
# INITIALIZE ATTRIBUTES
##########################################################

    def _initialize_attributes(self) -> None:
        """Initialize all attributes of the beam."""
        # Stirrups and shear attributes
        self._stirrup_s_l: PlainQuantity = 0 * cm
        self._stirrup_s_w: PlainQuantity = 0 * cm
        self._stirrup_s_max_l: PlainQuantity = 0 * cm
        self._stirrup_s_max_w: PlainQuantity = 0 * cm
        self._stirrup_n: int = 0
        self._A_v_min: PlainQuantity = 0 * cm**2/m
        self._A_v: PlainQuantity = 0 * cm**2 / m
        self._A_s_req_bot: PlainQuantity = 0 * cm**2
        self._A_s_req_top: PlainQuantity = 0 * cm**2
        self._A_v_req: PlainQuantity = 0 * cm**2 / m
        self._DCRv: float = 0
        self._DCRb_top: float = 0
        self._DCRb_bot: float = 0
        self._alpha: float = math.radians(90)

        # Design checks and effective heights
        self._rho_l_bot: PlainQuantity = 0 * dimensionless
        self._rho_l_top: PlainQuantity = 0 * dimensionless
        self._bot_rebar_centroid = 0 * mm
        self._top_rebar_centroid = 0 * mm
        self._c_d_top: float = 0
        self._c_d_bot: float = 0
        self._shear_checked = False  # Tracks if shear check or design has been done
        self._flexure_checked = False  # Tracks if shear check or design has been done

        # Initialize default concrete beam attributes
        self._initialize_code_attributes()
        # Longitudinal rebar attributes
        self._initialize_longitudinal_rebar_attributes()

        # Results attributes
        self._materials_shear: Dict = {}
        self._geometry_shear: Dict = {}
        self._forces_shear: Dict = {}
        self._all_shear_checks_passed: bool = False
        self._data_min_max_shear: Dict = {} 
        self._shear_reinforcement: Dict = {}
        self._shear_concrete: Dict = {}
        self._shear_all_checks: bool = False

        self._materials_flexure: Dict = {}
        self._geometry_flexure: Dict = {}
        self._forces_flexure: Dict = {}
        self._all_flexure_checks_passed: bool = False
        self._data_min_max_flexure: Dict = {}
        self._flexure_capacity_bot: Dict = {} 
        self._flexure_capacity_top: Dict = {}
        self._flexure_all_checks: bool = False

    def _initialize_code_attributes(self) -> None:
        if isinstance(self.concrete, Concrete_ACI_318_19):
            self._initialize_ACI_318_attributes()
        elif isinstance(self.concrete, Concrete_EN_1992_2004):
            self._initialize_EN_1992_2004_attributes()

    def _initialize_longitudinal_rebar_attributes(self) -> None:
        """Initialize all rebar-related attributes with default values."""
        # Bottom rebar defaults
        self._n2_b, self._d_b2_b = 0, 0 * mm
        self._n3_b, self._d_b3_b = 0, 0 * mm
        self._n4_b, self._d_b4_b = 0, 0 * mm

        # Top rebar defaults
        self._n2_t, self._d_b2_t = 0, 0 * mm
        self._n3_t, self._d_b3_t = 0, 0 * mm
        self._n4_t, self._d_b4_t = 0, 0 * mm

        # Unit system default starter rebar for initial effective height
        if self.concrete.unit_system == "metric":
            self._n1_b, self._d_b1_b = 2, 8 * mm
            self._n1_t, self._d_b1_t = 2, 8 * mm
        else:
            self._n1_b, self._d_b1_b = 2, 3/8 * inch
            self._n1_t, self._d_b1_t = 2, 3/8 * inch

        # Update dependent attributes
        self._update_longitudinal_rebar_attributes()

    def _initialize_ACI_318_attributes(self) -> None:
        if isinstance(self.concrete, Concrete_ACI_318_19):
            self._phi_V_n: PlainQuantity = 0*kN
            self._phi_V_s: PlainQuantity = 0*kN
            self._phi_V_c: PlainQuantity = 0*kN
            self._phi_V_max: PlainQuantity = 0*kN
            self._V_u: PlainQuantity = 0*kN
            self._M_u: PlainQuantity = 0*kNm
            self._M_u_bot: PlainQuantity = 0*kNm
            self._M_u_top: PlainQuantity = 0*kNm
            self._N_u: PlainQuantity = 0*kN
            self._A_cv: PlainQuantity = 0*cm**2
            self._k_c_min: PlainQuantity = 0*MPa
            self._sigma_Nu: PlainQuantity = 0*MPa
            self.V_c: PlainQuantity = 0*kN
            self._V_s_req: PlainQuantity = 0*kN
            self.phi_v: float = 0
            self.phi_t: float = 0
            self.lambda_factor = 0
            self._rho_w: PlainQuantity = 0*dimensionless
            self._lambda_s: float = 0
            self.f_yt: PlainQuantity = 0*MPa
            self._A_s_tension: PlainQuantity = 0*cm**2
            self._max_shear_ok: bool =False
            self._A_s_min_bot: PlainQuantity = 0*cm**2
            self._A_s_min_top: PlainQuantity = 0*cm**2
            self._A_s_max_bot: PlainQuantity = 0*cm**2
            self._A_s_max_top: PlainQuantity = 0*cm**2
            self._phi_M_n_bot: PlainQuantity = 0*kNm
            self._phi_M_n_top: PlainQuantity = 0*kNm
            self._d_b_max_bot: PlainQuantity = 0*mm
            self._d_b_max_top: PlainQuantity = 0*mm
            self.flexure_design_results_bot: DataFrame = None
            self.flexure_design_results_top: DataFrame = None

    def _initialize_EN_1992_2004_attributes(self) -> None:
        if isinstance(self.concrete, Concrete_EN_1992_2004):
            self._f_yk = self.steel_bar.f_y
            self._f_ck = self.concrete.f_ck
            self._f_ctm = self.concrete.f_ctm
            self._epsilon_cu3 = self.concrete._epsilon_cu3
            self._E_s = self.steel_bar.E_s
            self._V_Ed_1: PlainQuantity = 0*kN
            self._V_Ed_2: PlainQuantity = 0*kN
            self._N_Ed: PlainQuantity = 0*kN
            self._M_Ed: PlainQuantity = 0*kNm
            self._sigma_cd: PlainQuantity = 0*MPa
            self._V_Rd_c: PlainQuantity = 0*kN
            self._V_Rd_s: PlainQuantity = 0*kN
            self._V_s_req: PlainQuantity = 0*kN
            self._V_Rd_max: PlainQuantity = 0*kN
            self._V_Rd: PlainQuantity = 0*kN
            self._k_value: float = 0
            self._alpha_cc:float  = 0
            self._gamma_c: float = 0
            self._gamma_s:float  = 0
            self._f_ywk: PlainQuantity = 0*MPa
            self._f_ywd: PlainQuantity = 0*MPa
            self._f_yd: PlainQuantity = 0*MPa
            self._f_cd: PlainQuantity = 0*MPa
            self._A_s_tension: PlainQuantity = 0*cm**2
            self._A_p = 0*cm**2 # No prestressed for now
            self._sigma_cp: PlainQuantity = 0*MPa
            self._theta: float = 0
            self._cot_theta: float = 0
            self._z: PlainQuantity = 0*cm

##########################################################
# SET LONGITUDINAL AND TRANSVERSE REBAR AND UPDATE ATTRIBUTES
##########################################################

    def set_transverse_rebar(self, n_stirrups: int = 0, d_b:PlainQuantity = 0*mm, s_l:PlainQuantity = 0*cm) -> None:
        """Sets the transverse rebar in the beam section."""
        self._stirrup_n = n_stirrups
        self._stirrup_d_b = d_b
        self._stirrup_s_l = s_l
        n_legs = n_stirrups * 2
        A_db = (d_b ** 2) * math.pi / 4  # Area of one stirrup leg
        A_vs = n_legs * A_db  # Total area of stirrups
        self._A_v = A_vs / s_l  # Stirrup area per unit length

        # Update effective heights
        self._update_effective_heights()

    def set_longitudinal_rebar_bot(self, n1: int=0, d_b1: PlainQuantity=0*mm, n2: int = 0, d_b2: PlainQuantity=0*mm, 
                                n3: int=0, d_b3: PlainQuantity=0*mm, n4: int=0, d_b4: PlainQuantity=0*mm, 
                                ) -> None:
        """Update the bottom rebar configuration and recalculate attributes."""
        self._n1_b = n1 or self._n1_b
        self._d_b1_b = d_b1 or self._d_b1_b
        self._n2_b = n2 or self._n2_b
        self._d_b2_b = d_b2 or self._d_b2_b
        self._n3_b = n3 or self._n3_b
        self._d_b3_b = d_b3 or self._d_b3_b
        self._n4_b = n4 or self._n4_b
        self._d_b4_b = d_b4 or self._d_b4_b
        self._update_longitudinal_rebar_attributes()

    def set_longitudinal_rebar_top(self, n1: int, d_b1: PlainQuantity, n2: int=0, d_b2: PlainQuantity=0*mm, 
                                n3: int=0, d_b3: PlainQuantity = 0*mm, n4: int=0, d_b4: PlainQuantity=0*mm 
                                ) -> None:
        """Update the top rebar configuration and recalculate attributes."""
        self._n1_t = n1 or self._n1_t
        self._d_b1_t = d_b1 or self._d_b1_t
        self._n2_t = n2 or self._n2_t
        self._d_b2_t = d_b2 or self._d_b2_t
        self._n3_t = n3 or self._n3_t
        self._d_b3_t = d_b3 or self._d_b3_t
        self._n4_t = n4 or self._n4_t
        self._d_b4_t = d_b4 or self._d_b4_t
        self._update_longitudinal_rebar_attributes()
    
    def _calculate_longitudinal_rebar_area(self) -> None:
        """Calculate the total rebar area for a given configuration."""

        # AREA OF BOTTOM BARS
        self._A_s_bot = (self._n1_b * self._d_b1_b**2 * np.pi / 4 +
                self._n2_b * self._d_b2_b**2 * np.pi / 4 +
                self._n3_b * self._d_b3_b**2 * np.pi / 4 +
                self._n4_b * self._d_b4_b**2 * np.pi / 4)
    
        # AREA OF TOP BARS
        self._A_s_top = (self._n1_t * self._d_b1_t**2 * np.pi / 4 +
                self._n2_t * self._d_b2_t**2 * np.pi / 4 +
                self._n3_t * self._d_b3_t**2 * np.pi / 4 +
                self._n4_t * self._d_b4_t**2 * np.pi / 4)
        
    def _calculate_min_clear_spacing(self) -> None:
            """
            Calculates the maximum clear spacing between bars for the bottom rebar layers.

            Returns:
                PlainQuantity: The maximum clear spacing between bars in either the first or second layer.
            """

            def layer_clear_spacing(n_a: int, d_a: PlainQuantity, n_b: int, d_b: PlainQuantity) -> PlainQuantity:
                """
                Helper function to calculate clear spacing for a given layer.

                Parameters:
                    n_a (int): Number of bars in the first group of the layer.
                    d_a (PlainQuantity): Diameter of bars in the first group of the layer.
                    n_b (int): Number of bars in the second group of the layer.
                    d_b (PlainQuantity): Diameter of bars in the second group of the layer.

                Returns:
                    PlainQuantity: Clear spacing for the given layer.
                """
                effective_width = self.width - 2 * (self.c_c + self._stirrup_d_b)
                total_bars = n_a + n_b
                if total_bars <= 1:
                    return effective_width - max(d_a, d_b)  # Clear space for one bar
                total_bar_width = n_a * d_a + n_b * d_b
                return (effective_width - total_bar_width) / (total_bars - 1)

            # AVAIABLE CLEAR SPACING FOR BOTTOM BARS
            # Calculate clear spacing for each layer
            spacing_layer1_b = layer_clear_spacing(self._n1_b, self._d_b1_b, self._n2_b, self._d_b2_b)
            spacing_layer2_b = layer_clear_spacing(self._n3_b, self._d_b3_b, self._n4_b, self._d_b4_b)

            # Return the maximum clear spacing between the two layers
            self._available_s_bot=min(spacing_layer1_b, spacing_layer2_b)

            # AVAIABLE CLEAR SPACING FOR TOP BARS
            # Calculate clear spacing for each layer
            spacing_layer1_t = layer_clear_spacing(self._n1_t, self._d_b1_t, self._n2_t, self._d_b2_t)
            spacing_layer2_t = layer_clear_spacing(self._n3_t, self._d_b3_t, self._n4_t, self._d_b4_t)

            # Return the maximum clear spacing between the two layers
            self._available_s_top=min(spacing_layer1_t, spacing_layer2_t)

    def _calculate_long_rebar_centroid(self) -> None:
        """
        Calculates the centroid (baricenter) of a group of rebars based on their diameters, quantities, 
        and layer spacing.

        Returns:
            float: The calculated centroid height of the rebar group.
        """
        # BOTTOM BARS CENTROID
        # Calculate the vertical positions of the bar layers
        y1_b = self._d_b1_b / 2
        y2_b = self._d_b2_b / 2
        y3_b = max(self._d_b1_b, self._d_b2_b) + self.layers_spacing + self._d_b3_b / 2
        y4_b = max(self._d_b1_b, self._d_b2_b) + self.layers_spacing + self._d_b4_b / 2
        # Calculate the total area of each layer
        area_1_b = self._n1_b * self._d_b1_b**2*np.pi/4  # Area proportional to number of bars and their diameter
        area_2_b = self._n2_b * self._d_b2_b**2*np.pi/4
        area_3_b = self._n3_b * self._d_b3_b**2*np.pi/4
        area_4_b = self._n4_b * self._d_b4_b**2*np.pi/4

        # Calculate the centroid as a weighted average
        total_area_b = area_1_b + area_2_b + area_3_b + area_4_b
        if total_area_b == 0:
            return 0*mm  # Avoid division by zero if no bars are present

        self._bot_rebar_centroid = (area_1_b * y1_b + area_2_b * y2_b + area_3_b * y3_b + area_4_b * y4_b) / total_area_b
        
        #TOP BARS CENTROID
        # Calculate the vertical positions of the bar layers
        y1_t = self._d_b1_t / 2
        y2_t = self._d_b2_t / 2
        y3_t = max(self._d_b1_t, self._d_b2_t) + self.layers_spacing + self._d_b3_t / 2
        y4_t = max(self._d_b1_t, self._d_b2_t) + self.layers_spacing + self._d_b4_t / 2

        # Calculate the total area of each layer
        area_1_t = self._n1_t * self._d_b1_t**2*np.pi/4  # Area proportional to number of bars and their diameter
        area_2_t = self._n2_t * self._d_b2_t**2*np.pi/4
        area_3_t = self._n3_t * self._d_b3_t**2*np.pi/4
        area_4_t = self._n4_t * self._d_b4_t**2*np.pi/4

        # Calculate the centroid as a weighted average
        total_area_t = area_1_t + area_2_t + area_3_t + area_4_t
        if total_area_t == 0:
            return 0*mm  # Avoid division by zero if no bars are present

        self._top_rebar_centroid = (area_1_t * y1_t + area_2_t * y2_t + area_3_t * y3_t + area_4_t * y4_t) / total_area_t

    def _update_longitudinal_rebar_attributes(self) -> None:
        """Recalculate attributes dependent on rebar configuration for both top and bottom reinforcing."""
        self._calculate_longitudinal_rebar_area()
        self._calculate_long_rebar_centroid()
        self._calculate_min_clear_spacing()
        self._update_effective_heights()

    def _update_effective_heights(self) -> None:
        """Update effective heights and depths for moment and shear calculations."""
        self._c_mec_bot = self.c_c + self._stirrup_d_b + self._bot_rebar_centroid
        self._c_mec_top = self.c_c + self._stirrup_d_b + self._top_rebar_centroid
        self._d_bot = self._height - self._c_mec_bot
        self._d_top = self._height - self._c_mec_top
        # Use bottom or top effective height
        self._d_shear = min(self._d_bot, self._d_top)

##########################################################
# CHECK & DESIGN FLEXURE
##########################################################

    def design_flexure(self, forces: list[Forces]) -> DataFrame:
        """
        Designs flexural reinforcement for the beam using the provided forces and design code.
        Identifies the limiting cases for top and bottom reinforcement, designs for those cases, 
        and then checks flexural capacity for all forces.
        
        Returns
        -------
        DataFrame
            A DataFrame summarizing the flexural design results for all forces.
        """

        # Initialize limiting cases
        max_M_y_top = 0 * kN * m  # For negative M_y (top reinforcement design)
        max_M_y_bot = 0 * kN * m  # For positive M_y (bottom reinforcement design)

        # Identify the limiting cases
        for force in forces:
            # For top reinforcement, consider the minimum (most negative) moment
            if force._M_y <= max_M_y_top:
                max_M_y_top = force._M_y
                self._limiting_case_bot=force
            # For bottom reinforcement, consider the maximum positive moment
            if force._M_y > max_M_y_bot:
                max_M_y_bot = force._M_y
                self._limiting_case_top=force
        # Design flexural reinforcement for the limiting cases
        if self.concrete.design_code=="ACI 318-19" or self.concrete.design_code=="CIRSOC 201-25":
            _design_flexure_ACI_318_19(self, max_M_y_bot, max_M_y_top)
        else:
            raise ValueError(f"Longitudinal design method not implemented "
                            f"for concrete type: {type(self.concrete).__name__}")
        
        # Check flexural capacity for all forces with the assigned reinforcement
        all_results = self.check_flexure(forces)
        return all_results

    def check_flexure(self, forces: list[Forces]) -> DataFrame:
        
        # Initialize variables to track limiting cases
        max_dcr_top: float = 0
        max_dcr_bot: float = 0
        limiting_case_top = None
        limiting_case_bot = None
        limiting_case_top_details = None
        limiting_case_bot_details = None

        # To compile results for all forces
        self._flexure_results_list = []  # Store individual results for each force
        self._flexure_results_detailed_list = {}  # Store detailed results by force ID

        for force in forces:
            # Select the method based on design code
            if self.concrete.design_code=="ACI 318-19" or self.concrete.design_code=="CIRSOC 201-25":
                result = _check_flexure_ACI_318_19(self, force)
            # elif self.concrete.design_code=="EN 1992-2004":
            #     result =  _check_shear_EN_1992_2004(self, force)
            else:
                raise ValueError(f"Flexure design method not implemented for concrete type: {type(self.concrete).__name__}")  # noqa: E501
            self._flexure_results_list.append(result)

            # Store detailed results for this force
            self._flexure_results_detailed_list[force.id] = {
                'forces': self._forces_flexure.copy(),
                'min_max': self._data_min_max_flexure.copy(),
                'flexure_capacity_top': self._flexure_capacity_top.copy(),
                'flexure_capacity_bot': self._flexure_capacity_bot.copy(),
                'checks_pass': self._all_flexure_checks_passed
            }

            # Extract the DCR values for top and bottom from the results
            current_dcr_top = self._DCRb_top
            current_dcr_bot = self._DCRb_bot

            # Update top limiting case
            if current_dcr_top > max_dcr_top:
                max_dcr_top = current_dcr_top
                limiting_case_top = result
                limiting_case_top_details = self._flexure_results_detailed_list[force.id]

            # Update bottom limiting case
            if current_dcr_bot > max_dcr_bot:
                max_dcr_bot = current_dcr_bot
                limiting_case_bot = result
                limiting_case_bot_details = self._flexure_results_detailed_list[force.id]

        # Compile all results into a single DataFrame
        all_results = pd.concat(self._flexure_results_list, ignore_index=True)

        # Store limiting cases
        self._limiting_case_flexure_top = limiting_case_top
        self._limiting_case_flexure_bot = limiting_case_bot
        self._limiting_case_flexure_top_details = limiting_case_top_details
        self._limiting_case_flexure_bot_details = limiting_case_bot_details

        # Store maximum DCRs for easy access
        self._max_dcr_top = max_dcr_top
        self._max_dcr_bot = max_dcr_bot
        
        # Mark shear as checked
        self._flexure_checked = True
        return all_results

##########################################################
# CHECK & DESIGN SHEAR
##########################################################
     
    # Factory method to select the shear design method
    def design_shear(self, forces: list[Forces]) -> DataFrame:
        # Track the maximum A_v_req to identify the limiting case
        max_A_v_req = 0*cm**2/m 

        # Step 1: Identify the worst-case force
        for force in forces:
            if self.concrete.design_code=="ACI 318-19" or self.concrete.design_code=="CIRSOC 201-25":
                _design_shear_ACI_318_19(self, force)
            elif self.concrete.design_code=="EN 1992-2004":
                _design_shear_EN_1992_2004(self, force)
            else:
                raise ValueError(f"Shear design method not implemented for concrete type:"\
                    f"{type(self.concrete).__name__}")
            # Check if this result is the limiting case
            current_A_v_req = self._A_v_req
            if current_A_v_req >= max_A_v_req:
                max_A_v_req = current_A_v_req
                max_V_s_req = self._V_s_req

        # Step 2: Perform rebar design for the worst-case force
        section_rebar = Rebar(self)
        self.shear_design_results = section_rebar.transverse_rebar(max_A_v_req, max_V_s_req, self._alpha)
        self._best_rebar_design = section_rebar.transverse_rebar_design
        self._stirrup_s_l = self._best_rebar_design['s_l']
        self._stirrup_s_w = self._best_rebar_design['s_w']
        self._stirrup_s_max_l = self._best_rebar_design['s_max_l']
        self._stirrup_s_max_w = self._best_rebar_design['s_max_w']
        self.set_transverse_rebar(self._best_rebar_design['n_stir'],
                                  self._best_rebar_design['d_b'],
                                  self._best_rebar_design['s_l'])

        # Step 3: Check shear adequacy for all forces using the designed rebar
        all_results = self.check_shear(forces)  
        return all_results
    
    # Factory method to select the shear check method
    def check_shear(self, forces: list[Forces]) -> DataFrame:

        self._shear_results_list = []  # Store individual results for each force
        self._shear_results_detailed_list = {}  # Store detailed results by force ID
        max_dcr = 0  # Track the maximum DCR to identify the limiting case
        self._limiting_case_shear_details = None

        for force in forces:
            # Select the method based on design code
            if self.concrete.design_code=="ACI 318-19" or self.concrete.design_code=="CIRSOC 201-25":
                result = _check_shear_ACI_318_19(self, force)
            elif self.concrete.design_code=="EN 1992-2004":
                result =  _check_shear_EN_1992_2004(self, force)
            else:
                raise ValueError(f"Shear design method not implemented for concrete type: {type(self.concrete).__name__}")  # noqa: E501
            self._shear_results_list.append(result)
            self._shear_results_detailed_list[force.id] = {
                'forces': self._forces_shear.copy(),
                'shear_reinforcement': self._shear_reinforcement.copy(),
                'min_max': self._data_min_max_shear.copy(),
                'checks_pass': self._all_shear_checks_passed,
                'shear_concrete': self._shear_concrete.copy()
            }

            # Check if this result is the limiting case
            current_dcr = result['DCR'][0]
            if current_dcr > max_dcr:
                max_dcr = current_dcr
                self._limiting_case_shear = result
                self._limiting_case_shear_details = self._shear_results_detailed_list[force.id]

        # Compile all results into a single DataFrame
        all_results = pd.concat(self._shear_results_list, ignore_index=True)

        # Identify the most limiting case by Demand-to-Capacity Ratio (DCR) or other criteria
        self.limiting_case_shear = all_results.loc[all_results['DCR'].idxmax()]  # Select row with highest DCR

        self._shear_checked = True  # Mark shear as checked
        return all_results

##########################################################
# RESULTS
##########################################################
   
    # Beam results for Jupyter Notebook
    @property
    def data(self) -> None:
        markdown_content = f"Beam {self.label}, $b$={self.width.to('cm')}"\
                         f", $h$={self.height.to('cm')}, $c_{{c}}$={self.c_c.to('cm')}, \
                            Concrete {self.concrete.name}, Rebar {self.steel_bar.name}."
        self._md_data = markdown_content
        # Display the combined content
        display(Markdown(markdown_content))  # type: ignore

        return None
    
    @property
    def flexure_results(self) -> None:
        if not self._flexure_checked:
            warnings.warn("Flexural design has not been performed yet. Call _check_flexure() or "
                        "design_flexure() first.", UserWarning)
            self._md_flexure_results = "Flexural results are not available."
            return None
        # Check if limiting case details exist
        top_details = self._limiting_case_flexure_top_details or {}
        bot_details = self._limiting_case_flexure_bot_details or {}
        # Use limiting case results
        top_result_data = top_details.get('flexure_capacity_top')
        bot_result_data = bot_details.get('flexure_capacity_bot')

        checks_pass_top = top_details.get('checks_pass')
        checks_pass_bot = bot_details.get('checks_pass')
        warning_top = "⚠️ Some checks failed, see detailed results." if not checks_pass_top else ""
        warning_bot = "⚠️ Some checks failed, see detailed results." if not checks_pass_bot else ""

        # Pending for approval
        # all_checks_top = top_details.get('flexure_check')
        # all_checks_bot = bot_details.get('flexure_check')
        # all_checks = all_checks_top and all_checks_bot

        # Formatter instance for DCR formatting
        markdown_content = ""
        formatter = Formatter()
        
        # Handle top result data
        if top_result_data:
            top_rebar_1 = top_result_data['Value'][0]
            top_rebar_2 = top_result_data['Value'][1]
            area_top = top_result_data['Value'][7]
            Mu_top = self._limiting_case_flexure_top_details['forces']['Value'][0]
            Mn_top = top_result_data['Value'][9]
            DCR_top = top_result_data['Value'][10]

            rebar_top = f"{top_rebar_1}" + (f" ++ {top_rebar_2}" if top_rebar_2 != '-' else "")
            formatted_DCR_top = formatter.DCR(DCR_top)

            markdown_content += (
                f"Top longitudinal rebar: {rebar_top}, $A_{{s,top}}$ = {area_top} cm², $M_u$ = {Mu_top} kNm, "
                f"$\\phi M_n$ = {Mn_top} kNm → {formatted_DCR_top} {warning_top}\n\n"
            )
        else:
            markdown_content += "No top moment to check.\n\n"

        # Handle bottom result data
        if bot_result_data:
            bot_rebar_1 = bot_result_data['Value'][0]
            bot_rebar_2 = bot_result_data['Value'][1]
            area_bot = bot_result_data['Value'][7]
            Mu_bot = self._limiting_case_flexure_bot_details['forces']['Value'][1]
            Mn_bot = bot_result_data['Value'][9]
            DCR_bot = bot_result_data['Value'][10]

            rebar_bot = f"{bot_rebar_1}" + (f" ++ {bot_rebar_2}" if bot_rebar_2 != '-' else "")
            formatted_DCR_bot = formatter.DCR(DCR_bot)

            markdown_content += (
                f"Bottom longitudinal rebar: {rebar_bot}, $A_{{s,bot}}$ = {area_bot} cm², $M_u$ = {Mu_bot} kNm, "
                f"$\\phi M_n$ = {Mn_bot} kNm → {formatted_DCR_bot} {warning_bot}"
            )
        else:
            markdown_content += "No bottom moment to check."

        # markdown_content += 'Beam flexure checks PASS ✔️' if all_checks else "Beam flexure checks FAIL ❌"

        self._md_flexure_results = markdown_content
        display(Markdown(markdown_content))

    @property
    def shear_results(self) -> None:
        if not self._shear_checked:
            warnings.warn("Shear design has not been performed yet. Call check_shear() or "
                          "design_shear() first.", UserWarning)
            self._md_shear_results = "Shear results are not available."
            return None
        

        # Check if limiting case details exist
        shear_details = self._limiting_case_shear_details or {}
        # Use limiting case results
        limiting_reinforcement = shear_details.get('shear_reinforcement')
        limiting_forces = shear_details.get('forces')
        limiting_shear_concrete = shear_details.get('shear_concrete')
        checks_pass = shear_details.get('checks_pass')
        markdown_content = ""
        if shear_details:
            # Create FUFormatter instance and format FU value
            formatter = Formatter()
            formatted_DCR = formatter.DCR(limiting_shear_concrete['Value'][-1])
            if self._A_v == 0*cm:
                rebar_v = "not assigned"
            else:
                rebar_v = f"{int(limiting_reinforcement['Value'][0])}eØ{limiting_reinforcement['Value'][1]}/"\
                        f"{limiting_reinforcement['Value'][2]} cm"
            # Limitng cases checks 
            warning = "⚠️ Some checks failed, see detailed results." if not checks_pass else "" 
            markdown_content = f"Shear reinforcing {rebar_v}, $A_v$={limiting_reinforcement['Value'][6]} cm²/m"\
                            f", $V_u$={limiting_forces['Value'][1]} kN, $\\phi V_n$={limiting_shear_concrete['Value'][7]} kN → {formatted_DCR} {warning}"  # noqa: E501
        else:
            markdown_content += "No shear to check."
        self._md_shear_results = markdown_content
        display(Markdown(markdown_content))

        return None
    
    # Beam results for Jupyter Notebook
    @property
    def results(self) -> None:
        """
        Ensure that properties, flexure results, and shear results are available and display them.
        Handles cases where flexure or shear results are not yet available.
        """
        if not hasattr(self, '_md_properties'):
            self.data  # This will generate _md_properties
        if self._flexure_checked:
            self.flexure_results  # This will generate _md_flexure_results
        if self._shear_checked:
            self.shear_results  # This will generate _md_shear_results
        return None

    def flexure_results_detailed(self, force: Optional[Forces] = None) -> None:
        """
        Displays detailed flexure results.

        Parameters
        ----------
        forces : Forces, optional
            The specific Forces object to display results for. If None, displays results for the limiting case.
        Returns
        -------
        None
        """
        if not self._flexure_checked:
            warnings.warn("Flexural check has not been performed yet. Call _check_flexure or "
                        "design_flexure first.", UserWarning)
            self._md_flexure_results = "Flexure results are not available."
            return None

        # Determine which results to display (limiting case by default)
        if force:
            if force.id not in self._flexure_results_detailed_list:
                raise ValueError(f"No results found for Forces object with ID {force.id}.")
            result_data = self._flexure_results_detailed_list[force.id]
            top_result_data = result_data['flexure_capacity_top']
            bot_result_data = result_data['flexure_capacity_bot']
            forces_result = result_data['forces']
            min_max_result = result_data['min_max']
        else:
            if self._limiting_case_flexure_top_details is None:
                raise ValueError("Top limiting case details are not available.")
            if self._limiting_case_flexure_bot_details is None:
                raise ValueError("Bottom limiting case details are not available.")
            # Use the worst-case top and bottom scenarios
            top_result_data = self._limiting_case_flexure_top_details['flexure_capacity_top']
            bot_result_data = self._limiting_case_flexure_bot_details['flexure_capacity_bot']
            forces_result = {
                "Design forces": [
                    "Top max moment",
                    "Bottom max moment",
                ],
                "Variable": ["Mu,top", "Mu,bot"],
                "Value": [
                    round(self._limiting_case_flexure_top_details['forces']['Value'][0], 2),
                    round(self._limiting_case_flexure_bot_details['forces']['Value'][1], 2),
                ],
                "Unit": ["kNm", "kNm"]
            }
            min_max_result = {
            'Check': [
                'Min/Max As rebar top',
                'Minimum spacing top',
                'Min/Max As rebar bottom',
                'Minimum spacing bottom'
            ],
            'Unit': ['cm²', 'mm', 'cm²', 'mm'],
            'Value': [
                round(self._limiting_case_flexure_top_details['min_max']['Value'][0], 2),  # Top limiting case As
                round(self._limiting_case_flexure_top_details['min_max']['Value'][1], 2),  # Top limiting case spacing
                round(self._limiting_case_flexure_bot_details['min_max']['Value'][2], 2),  # Bottom limiting case As
                round(self._limiting_case_flexure_bot_details['min_max']['Value'][3], 2)   # Bottom limiting case spacing
            ],
            'Min.': [
                round(self._limiting_case_flexure_top_details['min_max']['Min.'][0], 2),  # Top limiting case As_min
                self._limiting_case_flexure_top_details['min_max']['Min.'][1],           # Top limiting case spacing min
                round(self._limiting_case_flexure_bot_details['min_max']['Min.'][2], 2),  # Bottom limiting case As_min
                self._limiting_case_flexure_bot_details['min_max']['Min.'][3]            # Bottom limiting case spacing min
            ],
            'Max.': [
                round(self._limiting_case_flexure_top_details['min_max']['Max.'][0], 2),  # Top limiting case As_max
                "",  # No max constraint for spacing
                round(self._limiting_case_flexure_bot_details['min_max']['Max.'][2], 2),  # Bottom limiting case As_max
                ""   # No max constraint for spacing
            ],
            'Ok?': [
                self._limiting_case_flexure_top_details['min_max']['Ok?'][0],  # Top As check
                self._limiting_case_flexure_top_details['min_max']['Ok?'][1],  # Top spacing check
                self._limiting_case_flexure_bot_details['min_max']['Ok?'][2],  # Bottom As check
                self._limiting_case_flexure_bot_details['min_max']['Ok?'][3]   # Bottom spacing check
            ]
            }

        # Create TablePrinter instances for detailed display
        print('===== BEAM FLEXURE DETAILED RESULTS =====')
        materials_printer = TablePrinter("MATERIALS")
        materials_printer.print_table_data(self._materials_flexure, headers='keys')

        geometry_printer = TablePrinter("GEOMETRY")
        geometry_printer.print_table_data(self._geometry_flexure, headers='keys')

        forces_printer = TablePrinter("FORCES")
        forces_printer.print_table_data(forces_result, headers='keys')

        min_max_printer = TablePrinter("MAX AND MIN LIMIT CHECKS")
        min_max_printer.print_table_data(min_max_result, headers='keys')

        capacity_printer = TablePrinter("FLEXURAL CAPACITY - TOP")
        capacity_printer.print_table_data(top_result_data, headers='keys')
        capacity_printer = TablePrinter("FLEXURAL CAPACITY - BOTTOM")
        capacity_printer.print_table_data(bot_result_data, headers='keys')

    def flexure_results_detailed_doc(self, force: Optional[Forces] = None) -> None:
        """
        Prints detailed flexure results in Word.

        Parameters
        ----------
        forces : Forces, optional
            The specific Forces object to display results for. If None, displays results for the limiting case.
        """
        if not self._flexure_checked:
            warnings.warn("Flexural check has not been performed yet. Call _check_flexure or "
                        "design_flexure first.", UserWarning)
            self._md_flexure_results = "Flexural results are not available."
            return None

        # Determine which results to display (limiting case by default)
        if force:
            if force.id not in self._flexure_results_detailed_list:
                raise ValueError(f"No results found for Forces object with ID {force.id}.")
            result_data = self._flexure_results_detailed_list[force.id]
            top_result_data = result_data['flexure_capacity_top']
            bot_result_data = result_data['flexure_capacity_bot']
            forces_result = result_data['forces']
            min_max_result = result_data['min_max']
        else:
            if self._limiting_case_flexure_top_details is None:
                raise ValueError("Top limiting case details are not available.")
            if self._limiting_case_flexure_bot_details is None:
                raise ValueError("Bottom limiting case details are not available.")
            # Use the worst-case top and bottom scenarios
            top_result_data = self._limiting_case_flexure_top_details['flexure_capacity_top']
            bot_result_data = self._limiting_case_flexure_bot_details['flexure_capacity_bot']
            forces_result = {
                "Design forces": [
                    "Top max moment",
                    "Bottom max moment",
                ],
                "Variable": ["Mu,top", "Mu,bot"],
                "Value": [
                    round(self._limiting_case_flexure_top_details['forces']['Value'][0], 2),
                    round(self._limiting_case_flexure_bot_details['forces']['Value'][1], 2),
                ],
                "Unit": ["kNm", "kNm"]
            }
            min_max_result = {
            'Check': [
                'Min/Max As rebar top',
                'Minimum spacing top',
                'Min/Max As rebar bottom',
                'Minimum spacing bottom'
            ],
            'Unit': ['cm²', 'mm', 'cm²', 'mm'],
            'Value': [
                round(self._limiting_case_flexure_top_details['min_max']['Value'][0], 2),  # Top limiting case As
                round(self._limiting_case_flexure_top_details['min_max']['Value'][1], 2),  # Top limiting case spacing
                round(self._limiting_case_flexure_bot_details['min_max']['Value'][2], 2),  # Bottom limiting case As
                round(self._limiting_case_flexure_bot_details['min_max']['Value'][3], 2)   # Bottom limiting case spacing
            ],
            'Min.': [
                round(self._limiting_case_flexure_top_details['min_max']['Min.'][0], 2),  # Top limiting case As_min
                self._limiting_case_flexure_top_details['min_max']['Min.'][1],           # Top limiting case spacing min
                round(self._limiting_case_flexure_bot_details['min_max']['Min.'][2], 2),  # Bottom limiting case As_min
                self._limiting_case_flexure_bot_details['min_max']['Min.'][3]            # Bottom limiting case spacing min
            ],
            'Max.': [
                round(self._limiting_case_flexure_top_details['min_max']['Max.'][0], 2),  # Top limiting case As_max
                "",  # No max constraint for spacing
                round(self._limiting_case_flexure_bot_details['min_max']['Max.'][2], 2),  # Bottom limiting case As_max
                ""   # No max constraint for spacing
            ],
            'Ok?': [
                self._limiting_case_flexure_top_details['min_max']['Ok?'][0],  # Top As check
                self._limiting_case_flexure_top_details['min_max']['Ok?'][1],  # Top spacing check
                self._limiting_case_flexure_bot_details['min_max']['Ok?'][2],  # Bottom As check
                self._limiting_case_flexure_bot_details['min_max']['Ok?'][3]   # Bottom spacing check
            ]
            }

        # Convert output Dicts into DataFrames
        df_materials = pd.DataFrame(self._materials_flexure)
        df_geometry = pd.DataFrame(self._geometry_flexure)
        df_forces = pd.DataFrame(forces_result)
        df_data_min_max = pd.DataFrame(min_max_result)
        df_flexure_capacity_top = pd.DataFrame(top_result_data)
        df_flexure_capacity_bottom = pd.DataFrame(bot_result_data)
        
        # Create a document builder instance
        doc_builder = DocumentBuilder(title='Concrete beam flexure check')

        # Add first section and table
        doc_builder.add_heading('Concrete beam flexure check', level=1)
        doc_builder.add_text(f'Design code: {self.concrete.design_code}')
        doc_builder.add_heading('Materials', level=2)
        doc_builder.add_table_data(df_materials)
        doc_builder.add_table_data(df_geometry)
        doc_builder.add_table_data(df_forces)

        # Add third section for limit checks
        doc_builder.add_heading('Limit Checks', level=2)
        doc_builder.add_table_data(df_data_min_max)

        # Add second section for flexural checks
        doc_builder.add_heading('Flexural Capacity Top', level=2)
        doc_builder.add_table_dcr(df_flexure_capacity_top)
        doc_builder.add_heading('Flexural Capacity Bottom', level=2)
        doc_builder.add_table_dcr(df_flexure_capacity_bottom)

        # Save the Word doc
        doc_builder.save(f"Concrete beam flexure check {self.concrete.design_code}.docx")
 
    def shear_results_detailed(self, force: Optional[Forces] = None) -> None:
        """
        Displays detailed shear results.

        Parameters
        ----------
        forces : Forces, optional
            The specific Forces object to display results for. If None, displays results for the limiting case.
        Returns
        -------
        None
        """
        if not self._shear_checked:
            warnings.warn("Shear check has not been performed yet. Call check_shear or " 
                          "design_shear first.", UserWarning)
            self._md_shear_results = "Shear results are not available."
            return None
         # Determine which results to display (limiting case by default)
        if force:
            force_id = force.id
            if force_id not in self._shear_results_detailed_list:
                raise ValueError(f"No results found for Forces object with ID {force_id}.")
            result_data = self._shear_results_detailed_list[force_id]
        else:
            # Default to limiting case
            result_data = self._limiting_case_shear_details

        # Create a TablePrinter instance and display tables
        print('===== BEAM SHEAR DETAILED RESULTS =====')
        materials_printer = TablePrinter("MATERIALS")
        materials_printer.print_table_data(self._materials_shear, headers='keys')
        geometry_printer = TablePrinter("GEOMETRY")
        geometry_printer.print_table_data(self._geometry_shear, headers='keys')
        forces_printer = TablePrinter("FORCES")
        forces_printer.print_table_data(result_data['forces'], headers='keys')
        steel_printer = TablePrinter("SHEAR STRENGTH")
        steel_printer.print_table_data(result_data['shear_reinforcement'], headers='keys')
        min_max_printer = TablePrinter("MAX AND MIN LIMIT CHECKS")
        min_max_printer.print_table_data(result_data['min_max'], headers='keys')
        concrete_printer = TablePrinter("CONCRETE STRENGTH")
        concrete_printer.print_table_data(result_data['shear_concrete'], headers='keys')

    def shear_results_detailed_doc(self, force: Optional[Forces] = None) -> None:
        """
        Prints detailed shear results in Word.

        Parameters
        ----------
        forces : Forces, optional
            The specific Forces object to display results for. If None, displays results for the limiting case.
        """
        if not self._shear_checked:
            warnings.warn("Shear check has not been performed yet. Call check_shear or " 
                          "design_shear first.", UserWarning)
            self._md_shear_results = "Shear results are not available."
            return None
         # Determine which results to display (limiting case by default)
        if force:
            force_id = force.id
            if force_id not in self._shear_results_detailed_list:
                raise ValueError(f"No results found for Forces object with ID {force_id}.")
            result_data = self._shear_results_detailed_list[force_id]
        else:
            # Default to limiting case
            result_data = self._limiting_case_shear_details
        
        # Convert output Dicts into DataFrames
        df_materials = pd.DataFrame(self._materials_shear)
        df_geometry = pd.DataFrame(self._geometry_shear)
        df_forces = pd.DataFrame(result_data['forces'])
        df_shear_reinforcement = pd.DataFrame(result_data['shear_reinforcement'])
        df_data_min_max = pd.DataFrame(result_data['min_max'])
        df_shear_concrete = pd.DataFrame(result_data['shear_concrete'])


        # Create a document builder instance
        doc_builder = DocumentBuilder(title='Concrete beam shear check')

        # Add first section and table
        doc_builder.add_heading('Concrete beam shear check', level=1)
        doc_builder.add_text(f'Design code: {self.concrete.design_code}')
        doc_builder.add_heading('Materials', level=2)
        doc_builder.add_table_data(df_materials)
        doc_builder.add_table_data(df_geometry)
        doc_builder.add_table_data(df_forces)

        # Add second section and another table (can use different data)
        doc_builder.add_heading('Limit checks', level=2)
        doc_builder.add_table_min_max(df_data_min_max)
        doc_builder.add_heading('Design checks', level=2)
        doc_builder.add_table_data(df_shear_reinforcement)
        doc_builder.add_table_dcr(df_shear_concrete)

        # Save the Word doc
        doc_builder.save(f"Concrete beam shear check {self.concrete.design_code}.docx")

    def _format_longitudinal_rebar_string(self, n1: int, d_b1: PlainQuantity, n2: int = 0, d_b2: PlainQuantity = 0*mm) -> str:
        """
        Returns a formatted string representing the rebars and their diameters.
        
        Example output:
            "2Ø10" (if only n1 and d_b1 are defined)
            "2Ø10+2Ø8" (if both n1/d_b1 and n2/d_b2 are defined)
        """
        if n1 == 0:
            rebar_string ="-"
        else:
            rebar_string = f"{n1}Ø{int(d_b1.magnitude)}"
            if n2 > 0 and d_b2.magnitude > 0:  # Check if n2 and d_b2 are defined
                rebar_string += f"+{n2}Ø{int(d_b2.magnitude)}"
        return rebar_string

##########################################################
# PLOT LONGIDTUINAL REBAR
##########################################################

    def plot(self) -> None:
        """
        Plots the longitudinal rebars for the beam or column.
        """
        # Call parent class to plot geometry
        super().plot()
        # Convert dimensions to consistent units (cm)
        width_cm: float = self.width.to('cm').magnitude
        height_cm: float = self.height.to('cm').magnitude
        c_c_cm: float = self.c_c.to('cm').magnitude
        stirrup_d_b_cm: float = self._stirrup_d_b.to('cm').magnitude
        layers_spacing_cm: float = self.layers_spacing.to('cm').magnitude

        # Calculate rebar positions
        # Bottom rebars
        self._plot_rebar_layer(width_cm, height_cm, c_c_cm, stirrup_d_b_cm, layers_spacing_cm,
                              self._n1_b, self._d_b1_b, self._n2_b, self._d_b2_b,max_db=self._d_b1_b, is_bottom=True)
        self._plot_rebar_layer(width_cm, height_cm, c_c_cm, stirrup_d_b_cm, layers_spacing_cm,
                              self._n3_b, self._d_b3_b, self._n4_b, self._d_b4_b, max_db=self._d_b1_b, is_bottom=True, is_second_layer=True)

        # Top rebars
        self._plot_rebar_layer(width_cm, height_cm, c_c_cm, stirrup_d_b_cm, layers_spacing_cm,
                              self._n1_t, self._d_b1_t, self._n2_t, self._d_b2_t, max_db=self._d_b1_t, is_bottom=False)
        self._plot_rebar_layer(width_cm, height_cm, c_c_cm, stirrup_d_b_cm, layers_spacing_cm,
                              self._n3_t, self._d_b3_t, self._n4_t, self._d_b4_t, max_db=self._d_b1_t, is_bottom=False, is_second_layer=True)

        # Show plot
        plt.show()

    def _plot_rebar_layer(self, width_cm: float, height_cm: float, c_c_cm: float, stirrup_d_b_cm: float, layers_spacing_cm: float,
                          n1: int, d_b1: PlainQuantity, n2: int, d_b2: PlainQuantity, max_db: PlainQuantity, 
                          is_bottom: bool=True, is_second_layer: bool=False) -> None:
        """
        Helper method to plot a single layer of rebars.
        """
        # Calculate y-position based on layer and bottom/top
        y_base = c_c_cm + stirrup_d_b_cm if is_bottom else height_cm - c_c_cm - stirrup_d_b_cm

        if is_second_layer:
            y_base += layers_spacing_cm + max_db.to('cm').magnitude if is_bottom else -layers_spacing_cm-max_db.to('cm').magnitude

        # Plot side bars (position 1 or 3)
        if n1 > 0:
            diameter_cm = d_b1.to('cm').magnitude
            y_center = y_base + diameter_cm/2 if is_bottom else y_base -diameter_cm/2
            for i in range(n1):
                x = c_c_cm + stirrup_d_b_cm + diameter_cm/2 + (i * (width_cm - 2 * (c_c_cm + stirrup_d_b_cm+diameter_cm/2)) / (n1 - 1))
                circle = Circle((x, y_center), diameter_cm / 2, color=CUSTOM_COLORS['dark_gray'], fill=True)
                self._ax.add_patch(circle)

        # Plot intermediate bars (position 2 or 4)
        if n2 > 0:
            diameter_cm = d_b2.to('cm').magnitude
            y_center = y_base + diameter_cm/2 if is_bottom else y_base -diameter_cm/2
            for i in range(n2):
                x = c_c_cm + stirrup_d_b_cm + diameter_cm/2 + ((i + 1) * (width_cm - 2 * (c_c_cm + stirrup_d_b_cm+diameter_cm/2)) / (n2 + 1))
                circle = Circle((x, y_center), diameter_cm / 2, color=CUSTOM_COLORS['dark_gray'], fill=True)
                self._ax.add_patch(circle)

##########################################################################################################


def clear_console() -> None:
    """
    Clears the console based on the operating system.
    """
    if os.name == 'nt':  # For Windows
        os.system('cls')
    else:  # For macOS and Linux
        os.system('clear')

def test_on_determine_nominal_moment_ACI_318_19()->None:
    concrete = Concrete_ACI_318_19(name="fc 4000", f_c=4000*psi)  
    steelBar = SteelBar(name="fy 60000", f_y=60*ksi)  
    custom_settings = {'clear_cover': 1.5*inch} 
    beam = RectangularBeam(
        label="B-12x24",
        concrete=concrete,
        steel_bar=steelBar,
        width=12*inch,  
        height=24*inch,
        settings=custom_settings  
    )
    
    beam.set_longitudinal_rebar_bot(n1=2,d_b1=1.128*inch, n2=1, d_b2=1.128*inch,  n3=2,d_b3=1*inch, n4=1, d_b4=1*inch)
    beam.set_longitudinal_rebar_top(n1=2,d_b1=1.128*inch, n2=1, d_b2=1.128*inch,  n3=2,d_b3=1*inch, n4=1, d_b4=1*inch)

    f = Forces(label='Test_01', M_y=400*kip*ft)
    beam._determine_nominal_moment_ACI_318_19(f)

def flexure_design_test() -> None:
    # clear_console()
    concrete= Concrete_ACI_318_19(name="C25",f_c=25*MPa) 
    steelBar= SteelBar(name="ADN 420", f_y=420*MPa)
    custom_settings = {'clear_cover': 2.5*cm}
    beam = RectangularBeam(label="101",
                                      concrete=concrete,steel_bar=steelBar,width=15*cm, height=50*cm,
                                       settings=custom_settings)
    f1 = Forces(label='C1', M_y=20*kNm)
    f2 = Forces(label='C2', M_y=-20*kNm)
    forces=[f1, f2]
    results=beam.design_flexure(forces)
    # print(beam.flexure_design_results_bot,'\n', beam.flexure_design_results_top)
    print(results)
    # beam.flexure_results_detailed()

def flexure_design_test_calcpad_example() -> None:
    concrete = Concrete_ACI_318_19(name="fc 4000", f_c=4000*psi)  
    steelBar = SteelBar(name="fy 60000", f_y=60*ksi)  
    custom_settings = {'clear_cover': 1.5*inch} 
    beam = RectangularBeam(
        label="B-12x24",
        concrete=concrete,
        steel_bar=steelBar,
        width=12*inch,  
        height=24*inch,
        settings=custom_settings  
    )
    
    #beam.set_longitudinal_rebar_bot(n1=2,d_b1=1.375*inch, n3=2,d_b3=1.27*inch)
    #beam.set_longitudinal_rebar_top(n1=2,d_b1=1.375*inch, n3=2,d_b3=1.27*inch)

    f = Forces(label='Test_01', V_z = 40*kip, M_y=400*kip*ft)
    f2 = Forces(label='Test_01', V_z = 100*kip, M_y=-400*kip*ft)
    forces=[f,f2]

    flexure_results = beam._design_flexure(forces)
    print(flexure_results)
 
def flexure_check_test() -> None:
    clear_console()
    concrete = Concrete_ACI_318_19(name="H-25",f_c=25*MPa) 
    steelBar = SteelBar(name="420", f_y=420*MPa) 

    beam = RectangularBeam(
        label="101",
        concrete=concrete,
        steel_bar=steelBar,
        width=20*cm,  
        height=60*cm,   
    )

    # beam.set_longitudinal_rebar_bot(n1=2,d_b1=20*mm)
    beam.set_longitudinal_rebar_bot(n1=2,d_b1=12*mm, n2=1, d_b2=12*mm, n3=2,d_b3=12*mm, n4=1, d_b4=10*mm)
    beam.set_longitudinal_rebar_top(n1=2,d_b1=16*mm)
    f1 = Forces(label='D', M_y=0*kNm, V_z=50*kN)
    f2 = Forces(label='L', M_y=-100*kNm)
    f3 = Forces(label='W', M_y=-50*kNm)
    f4 = Forces(label='S', M_y=110*kNm)
    forces=[f1,f2,f3,f4]
    print(beam.check_flexure(forces))
    
    # beam.check_shear()
    # beam.flexure_results_detailed()
    # beam.flexure_results_detailed_doc()
    # beam.shear_results_detailed_doc()

def flexure_Mn() -> None:
    clear_console()
    # MOMENTO NOMINAL SIMPLEMENTE ARMADO
    #Example from https://www.google.com/search?q=ACI+318+19+nominal+moment+of+section&oq=ACI+318+19+nominal+moment+of+section&gs_lcrp=EgZjaHJvbWUyBggAEEUYOTIHCAEQIRigAdIBCTIyNzkzajBqN6gCALACAA&sourceid=chrome&ie=UTF-8#fpstate=ive&vld=cid:dab5f5e7,vid:m4H0QbGDYIg,st:0
    concrete = Concrete_ACI_318_19(name="C6",f_c=6000*psi) 
    steelBar = SteelBar(name="G80", f_y=80000*psi) 
    beam = RectangularBeam(
        label="B20x30",
        concrete=concrete,
        steel_bar=steelBar,
        width=20 * inch,  
        height=30 * inch,  

    )
    A_s=10.92 * inch**2
    d=27*inch
    result=beam._determine_nominal_moment_simple_reinf_ACI_318_19(A_s,d)
    print(result.to(kip*ft))


    # MOMENTO NOMINAL DOBLEMENTE ARMADO
    concrete2 = Concrete_ACI_318_19(name="C4",f_c=4000*psi) 
    steelBar2 = SteelBar(name="G60", f_y=60000*psi) 
    beam2 = RectangularBeam(
        label="B14x27",
        concrete=concrete2,
        steel_bar=steelBar2,
        width=14 * inch,  
        height=27 * inch,   
    )
    A_s=6 * inch**2
    d=24*inch
    d_prime=2.5*inch
    A_s_prime=1.8*inch**2
    result=beam2._determine_nominal_moment_double_reinf_ACI_318_19(A_s, d, d_prime, A_s_prime)
    print(result.to(kip*ft))

def shear_ACI_metric() -> None:
    concrete= Concrete_ACI_318_19(name="C30",f_c=30*MPa) 
    steelBar= SteelBar(name="ADN 420", f_y=420*MPa)
    custom_settings = {'clear_cover': 30*mm}
    beam = RectangularBeam(label="101",
                                      concrete=concrete,steel_bar=steelBar,width=20*cm, height=50*cm,
                                       settings=custom_settings)
    f1 = Forces(label='1.4D', V_z=100*kN)
    f2 = Forces(label='1.2D+1.6L', V_z=1.55*kN)
    f3 = Forces(label='W', V_z=2.20*kN)
    f4 = Forces(label='S', V_z=8.0*kN)
    f5 = Forces(label='E', V_z=1.0*kN)
    forces=[f1, f2, f3, f4, f5]
    beam.set_longitudinal_rebar_bot(n1=2, d_b1=16 * mm)
    # beam.set_transverse_rebar(n_stirrups=1, d_b=6*mm, s_l=20*cm) 
    # results = beam.check_shear()
    results = beam._design_shear(forces)
    print(results)
    print(beam.shear_design_results)
    # beam.shear_results_detailed()
    # print(beam.shear_design_results)
    # print(beam.results)
    # beam.shear_results_detailed_doc()

def shear_ACI_imperial() -> None:
    concrete = Concrete_ACI_318_19(name="C4", f_c=4000*psi)  
    steelBar = SteelBar(name="ADN 420", f_y=60*ksi)  
    custom_settings = {'clear_cover': 1.5*inch} 
    beam = RectangularBeam(
        label="102",
        concrete=concrete,
        steel_bar=steelBar,
        width=10*inch,  
        height=16*inch,
        settings=custom_settings  
    )

    # f1 = Forces(label='D', V_z=37.727*kip, N_x=20*kip)
    f1 = Forces(label='D', V_z=37.727*kip)
    # f1 = Forces(label='D', V_z=8*kip)
    # f2 = Forces(label='L', V_z=6*kip) # No shear reinforcing
    forces=[f1]
    beam.set_transverse_rebar(n_stirrups=1, d_b=0.5*inch, s_l=6*inch)

    # beam.set_longitudinal_rebar_bot(n1=2, d_b1=0.625*inch)
    print(beam._A_v)
    results = beam._check_shear(forces)
    # results = beam.design_shear()
    print(results)
    # section.design_shear(f, A_s=0.847*inch**2)
    beam.shear_results_detailed()  
    # section.shear_results_detailed_doc()

def rebar() -> None:
    concrete= Concrete_ACI_318_19(name="H30",f_c=30*MPa) 
    steelBar= SteelBar(name="ADN 420", f_y=420*MPa)
    custom_settings = {'clear_cover': 30*mm, 'stirrup_diameter_ini':8*mm}
    section = RectangularBeam(
        label="V 20x50",
        concrete=concrete,
        steel_bar=steelBar,
        width=20*cm,  
        height=50*cm,
        settings=custom_settings  
    )
    as_req = 7*cm**2

    beam_rebar = Rebar(section)
    long_rebar_df = beam_rebar.longitudinal_rebar_ACI_318_19(A_s_req=as_req)
    best_design = beam_rebar.longitudinal_rebar_design
    print(long_rebar_df)

def rebar_df() -> None:
    concrete= Concrete_ACI_318_19(name="H30",f_c=30*MPa) 
    steelBar= SteelBar(name="ADN 420", f_y=420*MPa)
    custom_settings = {'clear_cover': 30*mm, 'stirrup_diameter_ini':8*mm}
    section = RectangularBeam(
        label="V 20x50",
        concrete=concrete,
        steel_bar=steelBar,
        width=20*cm,  
        height=50*cm,
        settings=custom_settings  
    )
    beam_rebar = Rebar(section)
    # Create a list of required steel areas from 0.5 to 10 with a step of 0.5 cm²
    as_req_list = np.arange(0.5, 10.5, 0.5)*cm**2
    # Initialize an empty DataFrame to store the results
    results_df = pd.DataFrame()

    # Loop through each required steel area
    for as_req in as_req_list:
        # Run the longitudinal_rebar_ACI_19 method
        long_rebar_df = beam_rebar.longitudinal_rebar_ACI_318_19(A_s_req=as_req)
        
        # Extract the first row of the resulting DataFrame
        first_row = long_rebar_df.iloc[0:1].copy()
        
        # Add a column to store the required steel area
        first_row['as_req'] = as_req.magnitude  # Store the magnitude (value without units)
        
        # Append the first row to the results DataFrame
        results_df = pd.concat([results_df, first_row], ignore_index=True)

    # Display the results DataFrame
    print(results_df)

if __name__ == "__main__":
    flexure_check_test()




import math
from pint.facets.plain import PlainQuantity
from typing import TYPE_CHECKING, Dict, Any, Tuple
import pandas as pd
from pandas import DataFrame

from mento.material import Concrete_EN_1992_2004
from mento.rebar import Rebar
from mento.units import MPa, mm, kNm, dimensionless, kN, inch, cm
from mento.forces import Forces  


if TYPE_CHECKING:
    from ..beam import RectangularBeam  # Import Beam for type checking only

def _initialize_variables_EN_1992_2004(self: 'RectangularBeam') -> None:
    """
    Initialize variables for EN 1992-2004 design code.
    """
    if isinstance(self.concrete, Concrete_EN_1992_2004):
        # Load settings for gamma factors
        self.settings.load_EN_1992_2004_settings()
        self._alpha_cc = self.settings.get_setting('alpha_cc')
        self._gamma_c = self.settings.get_setting('gamma_c')
        self._gamma_s = self.settings.get_setting('gamma_s')
        self._f_ywk = self._f_yk
        self._f_ywd = self._f_ywk/self._gamma_s
        self._f_cd = self._alpha_cc*self._f_ck/self._gamma_c
        self._f_yd = self._f_yk/self._gamma_s

##########################################################
# SHEAR CHECK AND DESIGN
##########################################################

def _initialize_shear_variables_EN_1992_2004(self: 'RectangularBeam', force: Forces) -> None:
    """
    Initialize variables for EN 1992-2004 design code.
    """
    if isinstance(self.concrete, Concrete_EN_1992_2004):
        # Set the initial variables
        self._N_Ed = force.N_x
        self._V_Ed_1 = force.V_z  # Consider the same shear at the edge of support and in d
        self._V_Ed_2 = force.V_z  # Consider the same shear at the edge of support and in d

        # Minimum shear reinforcement calculation
        rho_min = 0.08*math.sqrt(self._f_ck.to('MPa').magnitude) / (self._f_ywk)*MPa
        self._A_v_min = rho_min * self.width * math.sin(self._alpha)

        # Consider bottom or top tension reinforcement
        self._A_s_tension = self._A_s_bot if force._M_y >= 0*kNm else self._A_s_top

        # Compression stress, positive
        if force._M_y >= 0*kNm:
            self._rho_l_bot = min((self._A_s_tension + self._A_p) / (self.width * self._d_shear), 0.02*dimensionless)
        else:
            self._rho_l_top = min((self._A_s_tension + self._A_p) / (self.width * self._d_shear), 0.02*dimensionless)

        # Shear calculation for sections without rebar
        self._k_value = min(1 + math.sqrt(200 * mm / self._d_shear), 2)

        # Positive of compression
        self._sigma_cp = min(self._N_Ed / self.A_x, 0.2*self._f_cd)


def _shear_without_rebar_EN_1992_2004(self: 'RectangularBeam') -> PlainQuantity:
    self._stirrup_d_b = 0*mm
    self._theta = 0
    # Total shear capacity without rebar
    C_rdc = 0.18/self._gamma_c
    v_min = 0.035*self._k_value**(3/2)*math.sqrt(self._f_ck.to('MPa').magnitude)
    k_1 = 0.15
    V_Rd_c_min = ((v_min+k_1*self._sigma_cp.to('MPa').magnitude)* self.width * self._d_shear * MPa).to('kN')
    rho_l  = self._rho_l_bot if self._M_Ed >= 0*kNm else self._rho_l_top
    V_Rd_c = ((C_rdc*self._k_value*(100*rho_l*self._f_ck.to('MPa').magnitude)**(1/3)*MPa\
                +k_1*self._sigma_cp.to('MPa'))* self.width * self._d_shear).to('kN')        
    return max(V_Rd_c_min, V_Rd_c)
    
def _calculate_max_shear_strength_EN_1992_2004(self: 'RectangularBeam') -> None:
    """
    Calculate the maximum shear strength (V_Rd_max) and the corresponding strut angle (θ).
    """
    alpha_cw = 1  # Non-prestressed members or members subject to tensile stress due to axial force
    v_1 = 0.6 * (1 - self._f_ck.to('MPa').magnitude / 250)  # Strength reduction factor for concrete struts
    self._z = 0.9 * self._d_shear  # Lever arm

    # The θ angle is limited between 21.8° ≤ θ ≤ 45° (1 ≤ cot(θ) ≤ 2.5)
    # Check the minimum strut angle θ = 21.8° (cot(θ) = 2.5)
    theta_min: float = math.radians(21.8)
    cot_theta_min: float = 1 / math.tan(theta_min)

    V_Rd_max_min_angle = (alpha_cw * self.width * self._z * v_1 * self._f_cd / (cot_theta_min +
                                                                          math.tan(theta_min))).to('kN')

    if self._V_Ed_1 <= V_Rd_max_min_angle:
        # If within the minimum angle
        self._theta = theta_min
        self._cot_theta = cot_theta_min
        self._V_Rd_max = V_Rd_max_min_angle
        self._max_shear_ok = True
    else:
        # Check the maximum strut angle θ = 45° (cot(θ) = 1.0)
        theta_max: float = math.radians(45)
        cot_theta_max = 1 / math.tan(theta_max)
        V_Rd_max_max_angle: PlainQuantity = (alpha_cw * self.width * self._z * v_1 * self._f_cd / (cot_theta_max +
                                                                                            math.tan(theta_max))).to('kN')

        if self._V_Ed_1 > V_Rd_max_max_angle:
            self._theta = theta_max
            self._cot_theta = 1 / math.tan(self._theta)
            self._V_Rd_max = V_Rd_max_max_angle
            self._max_shear_ok = False
        else:
            self._max_shear_ok = True
            # Determine the angle θ of the strut based on the shear force
            self._theta = 0.5 * math.asin((self._V_Ed_1 / V_Rd_max_max_angle))
            self._cot_theta = 1 / math.tan(self._theta)
            self._V_Rd_max = (alpha_cw * self.width * self._z * v_1 * self._f_cd / (self._cot_theta +
                                                                              math.tan(self._theta))).to('kN')

def _calculate_required_shear_reinforcement_EN_1992_2004(self: 'RectangularBeam') -> None:
    """
    Calculate the required shear reinforcement area (A_v_req).
    """
    # Calculate the required shear reinforcement area
    self._A_v_req = max(
        (self._V_Ed_2 / (self._z * self._f_ywd * self._cot_theta)),  # Required area based on shear force
        self._A_v_min  # Minimum required area
    )
    self._V_Rd_s = (self._A_v * self._z * self._f_ywd * self._cot_theta)
    self._V_s_req = self._V_Rd_s
            #Maximum shear capacity is the same as the steel capacity
    self._V_Rd = self._V_Rd_s

def _check_shear_EN_1992_2004(self: 'RectangularBeam', force:Forces) -> DataFrame:
    if isinstance(self.concrete, Concrete_EN_1992_2004):
        # Initialize all the code related variables
        _initialize_variables_EN_1992_2004(self)
        _initialize_shear_variables_EN_1992_2004(self, force)            
        
        if self._stirrup_n == 0:
            # Calculate V_Rd_c
            self._V_Rd_c = _shear_without_rebar_EN_1992_2004(self)
            # According to EN1992-1-1 §6.2.1(4) minimum shear reinforcement should nevertheless be provided
            # according to EN1992-1-1 §9.2.2. The minimum shear reinforcement may be omitted in members where
            # transverse redistribution of loads is possible (such as slabs) and members of minor importance
            # which do not contribute significantly to the overall resistance and stability of the structure.
            self._A_v_req = self._A_v_min
            #Maximum shear capacity is the same as the concrete capacity
            self._V_Rd = self._V_Rd_c
            self._V_Rd_max = self._V_Rd
            self._max_shear_ok = self._V_Ed_1 <= self._V_Rd_max

        else:
            # Shear reinforcement calculations
            d_bs = self._stirrup_d_b
            s_l = self._stirrup_s_l
            n_legs = self._stirrup_n*2
            A_db = (d_bs ** 2) * math.pi / 4  # Area of one stirrup leg
            A_vs = n_legs * A_db  # Total area of stirrups
            self._A_v = A_vs / s_l  # Stirrup area per unit length
            
            # Calculate maximum shear strength
            _calculate_max_shear_strength_EN_1992_2004(self)

            # Calculate required shear reinforcement area
            _calculate_required_shear_reinforcement_EN_1992_2004(self)
        
            # Rebar spacing checks
            section_rebar = Rebar(self)
            n_legs_actual = self._stirrup_n * 2      # Ensure legs are even
            self._stirrup_s_l = self._stirrup_s_l
            self._stirrup_s_w = (self.width - 2 * self.c_c - self._stirrup_d_b) / (n_legs_actual - 1) 
            self._stirrup_s_max_l, self._stirrup_s_max_w = section_rebar.calculate_max_spacing_EN_1992_2004(self._alpha)

        self._DCRv = round(abs((self._V_Ed_2.to('kN').magnitude / self._V_Rd.to('kN').magnitude)),3)
        # Design results
        results = {
            'Label': self.label, #Beam label
            'Av,min': self._A_v_min.to('cm ** 2 / m'),  # Minimum shear reinforcement area
            'Av,req': self._A_v_req.to('cm ** 2 / m'), # Required shear reinforcing area
            'Av': self._A_v.to('cm ** 2 / m'),  # Provided stirrup reinforcement per unit length
            'VEd,1': self._V_Ed_1.to('kN'), # Max Vu for the design at the support
            'VEd,2': self._V_Ed_2.to('kN'), # Max Vu for the design at d from the support
            'VRd,c': self._V_Rd_c.to('kN'),  # Concrete contribution to shear capacity
            'VRd,s': self._V_Rd_s.to('kN'),  # Reinforcement contribution to shear capacity
            'VRd': self._V_Rd.to('kN'),  # Total shear capacity
            'VRd,max': self._V_Rd_max.to('kN'),  # Maximum shear capacity
            'VEd,1<VRd,max': self._max_shear_ok,  # Check if applied shear is within max shear capacity
            'VEd,2<VRd': self._V_Ed_2 <= self._V_Rd,  # Check if applied shear is within total capacity
            "DCR" :  self._DCRv
        }
        _initialize_dicts_EN_1992_2004_shear(self)
        return pd.DataFrame([results], index=[0])
    else:
        raise ValueError("Concrete type is not compatible with EN 1992 shear check.")

def _design_shear_EN_1992_2004(self: 'RectangularBeam', force:Forces) -> None:
    if isinstance(self.concrete, Concrete_EN_1992_2004):
        # Initialize all the code related variables
        _initialize_variables_EN_1992_2004(self)
        _initialize_shear_variables_EN_1992_2004(self, force) 
        # Calculate maximum shear strength
        _calculate_max_shear_strength_EN_1992_2004(self)
        # Calculate required shear reinforcement area
        _calculate_required_shear_reinforcement_EN_1992_2004(self)

        return None
    else:
        raise ValueError("Concrete type is not compatible with EN 1992 shear design.")

##########################################################
# FLEXURE CHECK AND DESIGN
##########################################################

def _min_max_flexural_reinforcement_ratio_EN_1992_2004(self: 'RectangularBeam') -> Tuple[float, float]:
    """
    Initialize variables for EN 1992-2004 design code.
    """
    # Calculate the minimum tensile reinforcement ratio
    rho_min = max(0.26 * self._f_ctm.to('MPa').magnitude / self._f_yk.to('MPa').magnitude, 0.0013)
    # Set the maximum tensile reinforcement ratio
    rho_max = 0.04

    # # For positive moments (tension in the bottom), set minimum reinforcement accordingly.
    # if force._M_y > 0 * kNm:
    #     rho_min_top = 0 * dimensionless
    #     rho_min_bot = rho_min
    # else:
    #     rho_min_top = rho_min
    #     rho_min_bot = 0 * dimensionless

    # Calculate minimum and maximum bottom reinforcement areas
    # self._A_s_min_bot = rho_min_bot * self._d_bot * self._width
    # self._A_s_max_bot = rho_max * self._d_bot * self._width

    return rho_min, rho_max

def _calculate_flexural_reinforcement_EN_1992_2004(self: 'RectangularBeam', M_Ed:PlainQuantity, d: PlainQuantity, d_prima: float) -> None:
    """
    Calculate the required top and bottom reinforcement areas for bending.
    """
    rho_min, rho_max = _min_max_flexural_reinforcement_ratio_EN_1992_2004(self)
    A_s_min = rho_min*d*self.width
    A_s_max = rho_max*d*self.width


    # Constants and material properties
    if isinstance(self.concrete, Concrete_EN_1992_2004):
        lambda_ = self.concrete.lambda_factor()  # Factor for effective compression zone depth (EN 1992-1-1)
        eta = self.concrete.eta_factor()  # Factor for concrete strength (EN 1992-1-1)

    # Design reinforcement yield strain
    epsilon_yd = (self._f_yd / self._E_s).to('dimensionless').magnitude  # Yield strain in reinforcement

    # Relative depth of compression zone at yielding of bottom reinforcement
    xi_eff_lim = lambda_ * (self._epsilon_cu3 / (self._epsilon_cu3 + epsilon_yd))

    # Compression zone depth limit
    x_eff_lim = xi_eff_lim * d

    # Limit moment for compressive reinforcement
    M_lim = (eta * self._f_cd * self.width * x_eff_lim * (d - 0.5 * x_eff_lim))

    # Check if compressive reinforcement is required
    if M_Ed <= M_lim:
        # Compressive reinforcement is NOT required
        print("Compressive reinforcement is NOT required.")

        # Relative design bending moment
        K = (M_Ed / (self.width * d**2 * eta * self._f_cd)).to('dimensionless').magnitude

        # Compression zone depth
        x_eff = d * (1 - math.sqrt(1 - 2 * K))

        # Lever arm of internal forces
        z = d - 0.5 * lambda_ * x_eff

        # Area of required tensile reinforcement
        A_s1 = (self.width * x_eff * eta * self._f_cd / self._f_yd).to('cm^2')

        # Ensure the area meets the minimum requirement
        A_s1 = max(A_s1, A_s_min)

        # No compressive reinforcement required
        A_s2 = 0 * cm**2

    else:
        # Compressive reinforcement is required
        print("Compressive reinforcement is required.")

        # Limit tensile reinforcement area
        A_s1_lim = (M_lim / ((d - 0.5 * x_eff_lim) * self._f_yd)).to('cm^2')

        # Extra moment to take with top reinforcement
        delta_M = M_Ed - M_lim

        # Required compressive reinforcement area
        A_s2 = (delta_M / ((d - d_prima) * self._f_yd)).to('cm^2')

        # Required tensile reinforcement area
        A_s1 = max(A_s1_lim + A_s2, A_s_min)


def _design_flexure_EN_1992_2004(self: 'RectangularBeam', max_M_y_bot: PlainQuantity, max_M_y_top: PlainQuantity) -> Dict[str, Any]:
    if isinstance(self.concrete, Concrete_EN_1992_2004):
        # Initialize all the code related variables
        _initialize_variables_EN_1992_2004(self)

        # Calculate reinforcement for the positive moment, even if it is 0
        (
            self._A_s_min_bot,
            self._A_s_max_bot,
            A_s_final_bot_Positive_M,
            A_s_comp_top,
            self._c_d_bot
        ) = _calculate_flexural_reinforcement_EN_1992_2004(self, max_M_y_bot, self._d_bot, self._c_mec_top)
        # Initialize bottom and top reinforcement
        self._A_s_bot = A_s_final_bot_Positive_M
        self._A_s_top = A_s_comp_top
                # If there is a negative moment, calculate the top reinforcement
        if max_M_y_top < 0:
            (
                self._A_s_min_top,
                self._A_s_max_top,
                A_s_final_top_Negative_M,
                A_s_comp_bot,
                self._c_d_top
            ) = _calculate_flexural_reinforcement_EN_1992_2004(self, abs(max_M_y_top), self._d_top, self._c_mec_bot)
            
            # Adjust reinforcement areas based on positive and negative moments
            self._A_s_bot = max(A_s_final_bot_Positive_M, A_s_comp_bot)                 
            self._A_s_top = max(A_s_comp_top,A_s_final_top_Negative_M)



        return None
    else:
        raise ValueError("Concrete type is not compatible with EN 1992 flexure design.")


##########################################################
# RESULTS
##########################################################

def _initialize_dicts_EN_1992_2004_shear(self: 'RectangularBeam') -> None:
    if isinstance(self.concrete, Concrete_EN_1992_2004):
        """Initialize the dictionaries used in check and design methods."""
        self._materials_shear = {
            "Materials": [
            "Section Label",
            "Concrete strength",
            "Steel reinforcement yield strength",
            "Safety factor for concrete",
            "Safety factor for steel",
            "Coefficient for long term effects and loading effects"
            ],
        "Variable": ["","fck", "fywk", "γc", "γs", "αcc"],
            "Value": [self.label, round(self.concrete.f_ck.to('MPa').magnitude,2),
                    round(self.steel_bar.f_y.to('MPa').magnitude,2),
                    self._gamma_c, self._gamma_s, self._alpha_cc
                    ],
            "Unit": ["", "MPa", "MPa", "","",""]
        }
        self._geometry_shear = {
                "Geometry": [
                "Section height",
                "Section width",
                "Clear cover",
                "Longitudinal tension rebar",
            ],
            "Variable": ["h", "b", "rgeom", "As"],
            "Value": [self.height.to('cm').magnitude, self.width.to('cm').magnitude, self.c_c.to('cm').magnitude,
                    round(self._A_s_tension.to('cm**2').magnitude,2)],
            "Unit": ["cm", "cm", "cm", "cm²"]
        }
        self._forces_shear = {
            "Design forces": [
                "Axial, positive for compression",
                "Shear",
            ],
            "Variable": ["NEd", "VEd,2"],
            "Value": [round(self._N_Ed.to('kN').magnitude,2), round(self._V_Ed_2.to('kN').magnitude,2)],
            "Unit": ["kN", "kN"]
        }
        # Min max lists
        if self._V_Rd_s == 0*kN:
            self._stirrup_d_b = 0*mm if self.concrete.unit_system == "metric" else 0 * inch  
        # Min max lists
        min_values = [None, None, self._A_v_min]   # Use None for items without a minimum constraint
        max_values = [self._stirrup_s_max_l, self._stirrup_s_max_w, None]  # Use None for items without a maximum constraint
        current_values = [self._stirrup_s_l, self._stirrup_s_w, self._A_v]  # Current values to check

        # Generate check marks based on the range conditions
        checks = [
            '✔️' if (min_val is None or curr >= min_val) and (max_val is None or curr <= max_val) else '❌'
            for curr, min_val, max_val in zip(current_values, min_values, max_values)
        ]
        self._all_shear_checks_passed = all(check == '✔️' for check in checks)
        self._data_min_max_shear = {
            'Check': ['Stirrup spacing along length', 'Stirrup spacing along width', 'Minimum shear reinforcement'],
            'Unit': ['cm', 'cm', 'cm²/m'],
            'Value': [round(self._stirrup_s_l.to('cm').magnitude,2), round(self._stirrup_s_w.to('cm').magnitude,2),
            round(self._A_v.to('cm**2/m').magnitude,2)],
            'Min.': ["", "", round(self._A_v_min.to('cm**2/m').magnitude,2)],
            'Max.': [round(self._stirrup_s_max_l.to('cm').magnitude,2), round(self._stirrup_s_max_w.to('cm').magnitude,2), ""],
            'Ok?': checks
        }
        self._shear_reinforcement = {
                "Shear reinforcement strength": [
                "Number of stirrups",
                "Stirrup diameter",
                "Stirrup spacing",
                "Effective height",
                "Minimum shear reinforcing",
                "Required shear reinforcing",
                "Defined shear reinforcing",
                "Shear rebar strength"
            ],
            "Variable": ["ns", "db", "s", "d", "Asw,min","Asw,req","Asw", "VRd,s"],
            "Value": [self._stirrup_n, self._stirrup_d_b.to('mm').magnitude, self._stirrup_s_l.to('cm').magnitude,
                    self._d_shear.to('cm').magnitude, round(self._A_v_min.to('cm**2/m').magnitude,2),
                    round(self._A_v_req.to('cm**2/m').magnitude,2),
                    round(self._A_v.to('cm**2/m').magnitude,2),
                    round(self._V_Rd_s.to('kN').magnitude,2)],
            "Unit": ["", "mm", "cm", "cm", "cm²/m","cm²/m", "cm²/m","kN"]
        }
        check_max = '✔️' if self._max_shear_ok else '❌'
        check_DCR = '✔️' if self._DCRv < 1 else '❌'
        rho_l  = self._rho_l_bot if self._M_Ed >= 0*kNm else self._rho_l_top
        self._shear_concrete = {
                "Shear strength": [
                "Longitudinal reinforcement ratio",
                'k value',
                "Axial stress",
                "Concrete strut angle",
                "Concrete strength",
                "Maximum shear strength",
                "Total shear strength", 
                "Max shear check",
                "Demand Capacity Ratio"
            ],
            "Variable": ["ρl", "k", "σcd","Θ","VRd,c", "VRd,max", "VRd","" ,"DCR"],
            "Value": [round(rho_l.magnitude,4),
                    round(self._k_value,2),round(self._sigma_cp.to('MPa').magnitude,2),
                    round(math.degrees(self._theta),1),
                    round(self._V_Rd_c.to('kN').magnitude,2), round(self._V_Rd_max.to('kN').magnitude,2), 
                    round(self._V_Rd.to('kN').magnitude,2), check_max, round(self._DCRv,2)],
            "Unit": ["", "", "MPa", "deg", "kN", "kN", "kN", "", check_DCR]
        }



# # *** (B), (C), (D), (E) ***
# def calculate_required_reinforcement_area(M_Ed, concrete, steel, geometry, reinforcement):
#     if M_Ed > 0:
#         # *** (B) ***
#         xi_eff_lim = concrete.lambda_() * (concrete.self._epsilon_cu3() / (concrete.self._epsilon_cu3() + steel.epsilon_yd()))
#         x_eff_lim = xi_eff_lim * geometry.d()
#         M_lim = concrete.eta_x_fcd() * geometry.width() * x_eff_lim * (geometry.d() - 0.5 * x_eff_lim)

#         if M_Ed <= M_lim:
#             # *** (D) ***
#             S_c_eff = M_Ed / (geometry.width() * (geometry.d() ** 2) * concrete.eta_x_fcd())
#             xi_eff = 1 - math.sqrt(1 - 2 * S_c_eff)
#             x_eff = xi_eff * geometry.d()
#             A_s = (concrete.eta_x_fcd() * x_eff * geometry.width()) / steel.f_yd()
#             A_s_req = max(A_s, reinforcement.A_s_min())
#             return ReinforcementAreas(A_s_req, 0)
#         else:
#             # *** (E) ***
#             a_2 = geometry.a()
#             A_s1 = M_lim / ((geometry.d() - 0.5 * x_eff_lim) * steel.f_yd())
#             delta_M = M_Ed - M_lim
#             A_s2 = delta_M / ((geometry.d() - a_2) * steel.f_yd())
#             A_s_req1 = max(A_s1 + A_s2, reinforcement.A_s_min())
#             return ReinforcementAreas(A_s_req1, A_s2)
#     else:
#         return ReinforcementAreas(0, 0)


# # *** (F), (G), (H), (I), (J) ***
# def calculate_provided_reinforcement_area(required_reinforcement_areas, reinforcement):
#     A_s_req1 = required_reinforcement_areas.bottom
#     A_s_req2 = required_reinforcement_areas.top
#     A_s_req = A_s_req1 + A_s_req2

#     n1 = math.ceil(A_s_req1 / reinforcement.A_phi())
#     A_s_prov1 = n1 * reinforcement.A_phi()

#     n2 = math.ceil(A_s_req2 / reinforcement.A_phi())
#     A_s_prov2 = n2 * reinforcement.A_phi()

#     A_s_prov = A_s_prov1 + A_s_prov2

#     if reinforcement.A_s_min() <= A_s_req <= reinforcement.A_s_max():
#         if A_s_prov <= reinforcement.A_s_max():
#             return ReinforcementAreas(A_s_prov1, A_s_prov2)
#         else:
#             return ReinforcementAreas(A_s_prov1, A_s_prov2, False)
#     else:
#         return ReinforcementAreas(A_s_prov1, A_s_prov2, False)


from typing import Optional, Dict, Any
from dataclasses import dataclass
from pint.facets.plain import PlainQuantity
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle, FancyBboxPatch


from mento.section import Section
from mento.material import SteelBar, Concrete
from mento.results import CUSTOM_COLORS
    
@dataclass
class RectangularSection(Section):
    
    label: Optional[str] = None

    def __init__(self, label: Optional[str], concrete: Concrete, steel_bar: SteelBar, width: PlainQuantity, height: PlainQuantity,
                 settings: Optional[Dict[str, Any]] = None) -> None:
        super().__init__(label, concrete, steel_bar, settings)

        if settings:
            self.settings.update(settings)  # Update with any provided settings

        self._width = width
        self._height = height
        self._A_x = self._width * self._height
        self._I_y = self._width*self._height**3/12
        self._I_z = self._height*self._width**3/12

        self._ax: Optional[plt.Axes] = None


    @property
    def width(self) -> PlainQuantity:
        "Beam width."
        return self._width.to('cm')

    @property
    def height(self) -> PlainQuantity:
        "Beam height."
        return self._height.to('cm')
    
    @property
    def A_x(self) -> PlainQuantity:
        "Cross section area."
        return self._A_x.to('cm**2')
    
    @property
    def I_y(self) -> PlainQuantity:
        "Moment of inertia about the Y axis."
        return self._I_y.to('cm**4')
    
    @property
    def I_z(self) -> PlainQuantity:
        "Moment of inertia about the Z axis."
        return self._I_z.to('cm**4')

    def plot(self) -> None:
        """
        Plots the rectangular section with a dark gray border, light gray hatch, and dimensions.
        Also plots the stirrup with rounded corners and thickness.
        """
        
        # Create figure and axis
        fig, self._ax = plt.subplots() 

        # Create a rectangle patch for the section
        rect = Rectangle((0, 0), self.width.magnitude, self.height.magnitude, 
                        linewidth=1.3, edgecolor=CUSTOM_COLORS['dark_gray'], facecolor=CUSTOM_COLORS['light_gray'])
        self._ax.add_patch(rect)

        # Calculate stirrup dimensions
        c_c = self.c_c.to('cm').magnitude
        stirrup_width = self.width.to('cm').magnitude - 2 * c_c
        stirrup_height = self.height.to('cm').magnitude - 2 * c_c
        stirrup_thickness = self._stirrup_d_b.to('cm').magnitude

        # Create rounded corners for the stirrup
        inner_radius = stirrup_thickness*2
        outer_radius = stirrup_thickness*3

        # Create the outer rounded rectangle for the stirrup
        outer_rounded_rect = FancyBboxPatch(
            (c_c, c_c),  # Bottom-left corner
            stirrup_width,  # Width
            stirrup_height,  # Height
            boxstyle=f'Round, pad=0, rounding_size={outer_radius}',  # Rounded corners
            edgecolor=CUSTOM_COLORS['dark_blue'],
            facecolor='white',
            linewidth=1
        )
        self._ax.add_patch(outer_rounded_rect)

        # Create the inner rounded rectangle for the stirrup (offset by thickness)
        inner_rounded_rect = FancyBboxPatch(
            (c_c + stirrup_thickness, c_c + stirrup_thickness),  # Bottom-left corner
            stirrup_width - 2 * stirrup_thickness,  # Width
            stirrup_height - 2 * stirrup_thickness,  # Height
            boxstyle=f'Round, pad=0, rounding_size={inner_radius}',  # Rounded corners
            edgecolor=CUSTOM_COLORS['dark_blue'],
            facecolor=CUSTOM_COLORS['light_gray'],
            linewidth=1
        )
        self._ax.add_patch(inner_rounded_rect)


        # Set plot limits with some padding
        padding = max(self.width.magnitude, self.height.magnitude) * 0.2
        self._ax.set_xlim(-padding, self.width.magnitude + padding)
        self._ax.set_ylim(-padding, self.height.magnitude + padding)

        # Text and dimension offsets
        dim_offset = 2.5
        text_offset = dim_offset + 2
        # Add width dimension
        self._ax.annotate(
            '',  # No text here, text is added separately
            xy=(0, -dim_offset),  # Start of arrow (left side)
            xytext=(self.width.magnitude, -dim_offset),  # End of arrow (right side)
            arrowprops={"arrowstyle": '<->', "lw": 1, "color": CUSTOM_COLORS['dark_blue']}
        )
        if self.concrete.unit_system == "imperial":
            width = '{:~P}'.format(self.width.to('inch'))
            height = '{:~P}'.format(self.height.to('inch'))
        else:
            width = '{:~P}'.format(self.width.to('cm'))
            height = '{:~P}'.format(self.height.to('cm'))
        # Add width dimension text below the arrow
        self._ax.text(
            self.width.magnitude / 2,  # Center of the arrow
            -text_offset,  # Slightly below the arrow
            width, 
            ha='center', 
            va='top',
            color =  CUSTOM_COLORS['dark_gray'],
        )

        # Add height dimension
        self._ax.annotate(
            '',  # No text here, text is added separately
            xy=(-dim_offset, 0),  # Start of arrow (bottom)
            xytext=(-dim_offset, self.height.magnitude),  # End of arrow (top)
            arrowprops={"arrowstyle": '<->', "lw": 1, "color": CUSTOM_COLORS['dark_blue']}
        )
        # Add height dimension text to the left of the arrow
        self._ax.text(
            -text_offset,  # Slightly to the left of the arrow
            self.height.magnitude / 2,  # Center of the arrow
            height, 
            ha='right', 
            va='center',
            color =  CUSTOM_COLORS['dark_gray'],
            rotation=90  # Rotate text vertically
        )

        # Set aspect of the plot to be equal
        self._ax.set_aspect('equal')
        # Remove axes for better visualization
        self._ax.axis('off')


import pytest
from mento.material import SteelBar, Concrete
from mento.units import MPa
from mento.section import Section     # Assuming this is the module name
from mento.settings import Settings
from typing import Dict, Any

@pytest.fixture()
def setup_basic_section() -> Section:
    concrete = Concrete('C25')
    steel_bar = SteelBar(name='B500S', f_y=500 * MPa)
    section = Section(label="V101",concrete=concrete, steel_bar=steel_bar)
    return section


def test_section_id_increment(setup_basic_section: Section) -> None:
    section1 = setup_basic_section
    section2 = Section(label="V102",concrete=section1.concrete, steel_bar=section1.steel_bar)
    assert section2.id == section1.id + 1


def test_initial_settings(setup_basic_section: Section) -> None:
    section = setup_basic_section
    assert isinstance(section.settings, Settings)


def test_update_settings(setup_basic_section: Section) -> None:
    section = setup_basic_section
    new_settings: Dict[str, Any] = {'clear_cover': 20}
    section.update_settings(new_settings)
    assert section.settings.get_setting('clear_cover') == 20


def test_get_settings(setup_basic_section: Section) -> None:
    section = setup_basic_section
    settings = section.get_settings
    assert isinstance(settings, dict)
    assert 'clear_cover' in settings



# Expose units and Quantity directly to the user
from .units import ureg, m, cm, mm, kN, kNm, MPa, GPa, kg, sec, psi, lb, kip, ksi, inch, ft, deg  # noqa: F401

# Export Quantity for user convenience
Quantity = ureg.Quantity

# Expose classes from different modules
from .node import Node # noqa: F401, E402
from .forces import Forces # noqa: F401, E402
from .material import Concrete_ACI_318_19, SteelBar, Concrete_CIRSOC_201_25, Concrete_EN_1992_2004 # noqa: F401, E402
from .beam import RectangularBeam # noqa: F401, E402
from .results import Formatter, TablePrinter, DocumentBuilder # noqa: F401, E402
from .codes import EN_1992_2004_beam, ACI_318_19_beam # noqa: F401, E402
from .summary import BeamSummary # noqa: F401, E402


from typing import Optional, Dict, Any

from mento.units import mm, inch
from mento.material import Concrete

class Settings:
    default_settings_metric: Dict[str, Any] = {
            # Beam design settings
            'clear_cover': 25 * mm, 
            'clear_spacing': 25 * mm, 
            'stirrup_diameter_ini': 8 * mm,
            'vibrator_size': 30 * mm, 
            'layers_spacing': 25 * mm,
            'max_diameter_diff': 5*mm,
            'max_bars_per_layer': 5,
            'minimum_longitudinal_diameter': 8*mm,
        }
    default_settings_imperial: Dict[str, Any] = {
            # Beam design settings
            'clear_cover': 1 * inch, 
            'clear_spacing': 1 * inch, 
            'stirrup_diameter_ini': 3/8*inch,
            'vibrator_size': 1.25*inch, 
            'layers_spacing': 1*inch,
            'max_diameter_diff': 2/8*inch,
            'max_bars_per_layer': 5,
            'minimum_longitudinal_diameter': 3/8*inch,
        }
    
    def __init__(self, concrete: Optional[Concrete] = None, settings: Optional[Dict[str, Any]] = None):
        """
        Initialize the Settings class with defaults based on unit system.
        
        Parameters
        ----------
        concrete : Concrete, optional
            A Concrete object to determine the unit system.
        settings : dict, optional
            Custom settings to override defaults.
        """
        # Select default settings based on the concrete's unit system
        if concrete and concrete.unit_system == "imperial":
            self.settings = self.default_settings_imperial.copy()
        else:  # Default to metric
            self.settings = self.default_settings_metric.copy()
        
        # Update defaults with provided settings if any
        if settings:
            self.update(settings)

        self.ACI_318_19_settings : Dict[str, Any] = {
                'lambda': 1, # Normalweight concrete
                'phi_v': 0.75, # Shear strength reduction factor
                'phi_c': 0.65, # Compression controlled strength reduction factor
                'phi_t': 0.90,  # Tension controlled strength reduction factor
                'flexural_min_reduction': "False"  # True selects 4/3 of calculated steel if it's less than minimum
            }
        self.EN_1992_2004_settings : Dict[str, Any] = {
                'gamma_c': 1.5,
                'gamma_s': 1.15, 
                'alpha_cc': 1.00, 
            }
    def load_ACI_318_19_settings(self) -> None:
        """
        Load settings specific to ACI 318-19.
        This will override only the settings that are different in ACI 318-19.
        """
        # Update current settings with ACI 318-19 specific settings
        self.add_settings(self.ACI_318_19_settings)

    def load_EN_1992_2004_settings(self) -> None:
        """
        Load settings specific to ACI 318-19.
        This will override only the settings that are different in ACI 318-19.
        """
        # Update current settings with ACI 318-19 specific settings
        self.add_settings(self.EN_1992_2004_settings)
    
    def get_setting(self, key: str) -> Any:
        if key in self.settings:
            return self.settings.get(key)
        else:
            raise KeyError(f"Setting '{key}' does not exist.")

    def set_setting(self, key: str, value: Any) -> None:
        if key in self.settings:
            self.settings[key] = value
        else:
            raise KeyError(f"Setting '{key}' does not exist.")
        
    def add_settings(self, new_settings: Dict[str, Any]) -> None:
            """Adds the settings dictionary to the current settings."""
            for key, value in new_settings.items():
                self.settings[key] = value
    
    def update(self, new_settings: Dict[str, Any]) -> None:
        """Updates the settings dictionary with new values."""
        self.add_settings(new_settings)

    def __str__(self) -> str:
        """Customize the string representation for user-friendly display."""
        output = "Settings:\n"
        for key, value in self.settings.items():
            output += f"  {key}: {value}\n"
        return output.strip()



from __future__ import annotations
from typing import Any, Dict, TYPE_CHECKING, Tuple
import math
import pandas as pd
import numpy as np

from mento.units import psi, mm, cm, inch, MPa

if TYPE_CHECKING:
    from mento.beam import RectangularBeam
    from pint.facets.plain import PlainQuantity
    from pandas import DataFrame

class Rebar:
    def __init__(self, beam: RectangularBeam):
        """
        Initializes the Rebar object with the associated beam and settings.
        """
        self.beam = beam
        self.min_clear_spacing = self.beam.settings.get_setting('clear_spacing')
        self.vibrator_size = self.beam.settings.get_setting('vibrator_size')
        self.max_diameter_diff = self.beam.settings.get_setting('max_diameter_diff')
        self.max_bars_per_layer = self.beam.settings.get_setting('max_bars_per_layer')
        self.min_long_rebar = self.beam.settings.get_setting('minimum_longitudinal_diameter')
        self._long_combos_df: DataFrame = None
        self._trans_combos_df: DataFrame = None
        self._clear_spacing: PlainQuantity = 0*mm
        # Unit system default rebar
        if self.beam.concrete.unit_system == "metric":
            self.rebar_diameters = [6*mm, 8*mm, 10*mm, 12*mm, 16*mm, 20*mm, 25*mm, 32*mm]
            self.rebar_areas = {d: (math.pi * d ** 2) / 4 for d in self.rebar_diameters}
        else:
            self.rebar_diameters = [3*inch/8, 4*inch/8, 5*inch/8, 6*inch/8, 7*inch/8, 8*inch/8, 1.128*inch,
                                    1.27*inch, 1.41*inch, 1.693*inch]
            rebar_areas_list = [(d**2 * np.pi / 4) for d in self.rebar_diameters]
            self.rebar_areas = dict(zip(self.rebar_diameters, rebar_areas_list))

    @property
    def longitudinal_rebar_design(self) -> DataFrame:
        return self._long_combos_df.iloc[0]
    
    @property
    def transverse_rebar_design(self) -> DataFrame:
        return self._trans_combos_df.iloc[0]

##########################################################
# TRANSVERSE REBAR DESIGN
##########################################################

    def calculate_max_spacing_ACI_318_19(self, V_s_req: PlainQuantity,
                                   A_cv: PlainQuantity) -> Tuple[PlainQuantity, PlainQuantity]:
        """
        Calculate the maximum allowable spacing across the length and width of the beam
        based on design requirements.

        Parameters:
        -----------
        V_s_req : float
            Required shear force for the rebar.
        A_cv : float
            Effective shear area of the concrete section.

        Returns:
        --------
        tuple
            (s_max_l, s_max_w): The maximum spacing across the length and width of the beam.
        """

        f_c=self.beam.concrete.f_c
        lambda_factor = self.beam.settings.get_setting('lambda')
        
        # Determine maximum spacing based on V_s_req condition
        # Maximum spacing for lower shear demand
        if self.beam.concrete.unit_system == "metric":
            if V_s_req <= 0.083 * lambda_factor * math.sqrt(f_c / MPa) * MPa * A_cv:
                s_max_l = min(self.beam._d_shear / 2, 60 * cm)
                s_max_w = min(self.beam._d_shear, 60 * cm)
            else:
            # Maximum spacing for higher shear demand
                s_max_l = min(self.beam._d_shear / 4, 30 * cm)
                s_max_w = min(self.beam._d_shear / 2, 30 * cm)
        else:
            if V_s_req <= 4 * lambda_factor * math.sqrt(f_c / psi) * psi * A_cv:
                s_max_l = min(self.beam._d_shear / 2, 24*inch)
                s_max_w = min(self.beam._d_shear, 24*inch)
            else:
                s_max_l = min(self.beam._d_shear / 4, 12*inch)
                s_max_w = min(self.beam._d_shear / 2, 12*inch)
        
        return s_max_l, s_max_w  
    
    def calculate_max_spacing_EN_1992_2004(self, alpha: float) -> Tuple[PlainQuantity, PlainQuantity]:
        """
        Calculate the maximum allowable spacing across the length and width of the beam
        based on design requirements for EN 1992-2004.
        
        Parameters:
        -----------
        alpha: stirrups angle

        Returns:
        --------
        tuple
            (s_max_l, s_max_w): The maximum spacing along the length and width of the beam.
        """
        s_max_l = 0.75*self.beam._d_shear*(1+1 / math.tan(alpha))
        s_max_w = min(0.75*self.beam._d_shear, 60 * cm)
           
        return s_max_l, s_max_w

    def transverse_rebar_ACI_318_19(self, V_s_req: PlainQuantity) -> Any:

        if self.beam.concrete.unit_system == "metric":
            valid_diameters = self.rebar_diameters[2:5] #Minimum 10 mm
        else: 
            valid_diameters = self.rebar_diameters[0:3]
        
        A_cv = self.beam.width*self.beam._d_shear
        s_max_l, s_max_w = self.calculate_max_spacing_ACI_318_19(V_s_req, A_cv)

        return valid_diameters, s_max_l, s_max_w
    
    def transverse_rebar_CIRSOC_201_25(self, V_s_req: PlainQuantity) -> Any:

        valid_diameters = self.rebar_diameters[0:5] #Minimum 6 mm
        
        A_cv = self.beam.width*self.beam._d_shear
        s_max_l, s_max_w = self.calculate_max_spacing_ACI_318_19(V_s_req, A_cv)

        return valid_diameters, s_max_l, s_max_w
    
    def transverse_rebar_EN_1992_2004(self, alpha: float) -> Any:
        valid_diameters = self.rebar_diameters[0:5] #Minimum 6 mm

        s_max_l, s_max_w = self.calculate_max_spacing_EN_1992_2004(alpha)

        return valid_diameters, s_max_l, s_max_w

    def transverse_rebar(self, A_v_req: PlainQuantity, V_s_req: PlainQuantity, alpha:float) -> DataFrame:
        """
        Computes the required transverse reinforcement based on ACI 318-19.

        Args:
            A_v_req: Required area for transverse reinforcement.

        Returns:
            A dictionary containing the transverse rebar design.
        """

        # Prepare the list for valid combinations
        valid_combinations = []

        # Get code specific limitations
        if self.beam.concrete.design_code=="ACI 318-19":
            valid_diameters, s_max_l, s_max_w = self.transverse_rebar_ACI_318_19(V_s_req)
        elif self.beam.concrete.design_code=="CIRSOC 201-25":
            valid_diameters, s_max_l, s_max_w = self.transverse_rebar_CIRSOC_201_25(V_s_req)
        elif self.beam.concrete.design_code == "EN 1992-2004":
            valid_diameters, s_max_l, s_max_w = self.transverse_rebar_EN_1992_2004(alpha)
        else:
            raise ValueError(f"Shear design method not implemented for concrete type: {type(self.beam.concrete).__name__}")

        # Iterate through available diameters
        for d_b in valid_diameters:
            # Start with 2 legs = 1 stirrup
            n_legs = 2

            # Start with maximum allowed spacing s_max_l
            if self.beam.concrete.unit_system == "metric":
                s_l: PlainQuantity = math.floor(s_max_l.to('cm').magnitude)*cm
            else:
                s_l = math.floor(s_max_l.to('inch').magnitude)*inch

            while True:
                # Calculate spacing based on current legs
                n_stirrups = math.ceil(n_legs / 2)  # Number of stirrups based on number of legs
                n_legs_actual = n_stirrups * 2      # Ensure legs are even
                # Consider 1 leg less for spacing laong width
                s_w = (self.beam.width - 2 * self.beam.c_c - self.beam._stirrup_d_b) / (n_legs_actual - 1)  # noqa: F841

                A_db = self.rebar_areas[d_b]  # Area of a stirrup bar
                A_vs = n_legs_actual * A_db  # Area of vertical stirrups
                A_v: PlainQuantity = A_vs / s_l  # Area of vertical stirrups per unit length

                # Store the valid combination if spacing is also valid
                if self.beam.concrete.unit_system == "metric":
                    # Check if the calculated A_v meets or exceeds the required A_v
                    if A_v >= A_v_req:
                        valid_combinations.append({
                            'n_stir': int(n_stirrups),
                            'd_b': d_b,
                            's_l': s_l.to('cm'),  # spacing along length
                            's_w': s_w.to('cm'), # spacing along width
                            'A_v': A_v.to('cm**2/m'),
                            's_max_l': s_max_l.to('cm'),
                            's_max_w': s_max_w.to('cm')
                        })
                        # Stop checking larger diameters
                        break  
                        
                    # If A_v is insufficient, reduce s_l by 1 cm
                    s_l -= 1 * cm
                    # If s_l becomes less than 5 cm, increase the number of legs by 2 and reset s_l to s_max_l
                    if s_l < 5 * cm:  # If spacing is less than 5 cm, increase 1 stirrup
                        n_legs += 2
                        s_l = math.floor(s_max_l.to('cm').magnitude)*cm # Reset s_l to the max allowed spacing
                    # Break the loop if legs exceed the limit of legs or max stirrup diameter
                    if n_legs > 6:
                        break
                else:
                    # Check if the calculated A_v meets or exceeds the required A_v
                    if A_v >= A_v_req:
                        valid_combinations.append({
                            'n_stir': int(n_stirrups),
                            'd_b': d_b,
                            's_l': s_l.to('inch'),  # spacing along length
                            's_w': s_w.to('inch'), # spacing along width
                            'A_v': A_v.to('inch**2/ft'),
                            's_max_l': s_max_l.to('inch'),
                            's_max_w': s_max_w.to('inch')
                        })
                        # Stop checking larger diameters
                        break  
                        
                    # If A_v is insufficient, reduce s_l by 1 inch
                    s_l -= 1 * inch
                    # If s_l becomes less than 2 inch, increase the number of legs by 2 and reset s_l to s_max_l
                    if s_l < 2 * inch:  # If spacing is less than 2 inch, increase 1 stirrup
                        n_legs += 2
                        s_l = math.floor(s_max_l.to('inch').magnitude)*inch # Reset s_l to the max allowed spacing
                    # Break the loop if legs exceed the limit of legs or max stirrup diameter
                    if n_legs > 6:
                        break

        # Create a DataFrame with all valid combinations
        df_combinations = pd.DataFrame(valid_combinations)

        # Sort combinations by the total rebar area required (ascending)
        # Sort by 'A_v' first, then by 'n_stir' to prioritize fewer bars
        df_combinations.sort_values(by=['n_stir','A_v'], inplace=True)
        df_combinations.reset_index(drop=True, inplace=True)
        self._trans_combos_df = df_combinations
        return df_combinations

##########################################################
# LONGITUDINAL REBAR DESIGN
##########################################################

    def longitudinal_rebar_ACI_318_19(self, A_s_req: PlainQuantity) -> DataFrame:
        """
        Computes the required longitudinal reinforcement based on ACI 318-19.
        
        Args:
            A_s_req: Required longitudinal rebar area.

        Returns:
            A DataFrame containing the best combinations of rebar details.
            If no combination satisfies A_s_req, returns the combination with the maximum possible area.
        """
        self.A_s_req = A_s_req
        effective_width = self.beam.width - 2 * (self.beam.c_c + self.beam._stirrup_d_b)

        # Variables to track the combinations
        valid_combinations = []
        best_fallback_combination = None  # To store the best fallback design
        max_fallback_area = 0 * cm**2  # To track the maximum area in fallback cases
        # Create a list of rebar diameters that are equal to or greater than the minimum diameter
        if self.beam.concrete.unit_system == "metric":
            self.min_long_rebar = 10*mm
        else: 
            self.min_long_rebar = 3*inch/8
        valid_rebar_diameters = [d for d in self.rebar_diameters if d >= self.min_long_rebar]

        for d_b1 in valid_rebar_diameters: # Without taking Ø6 as a possible solution
                for d_b2 in [d for d in valid_rebar_diameters if d <= d_b1]:
                    for d_b3 in [d for d in valid_rebar_diameters if d <= d_b2]:
                        for d_b4 in [d for d in valid_rebar_diameters if d <= d_b3]:

                            # Condition 5: |d_b1 - d_b2| and |d_b3 - d_b4| must not exceed max_diameter_diff
                            # Ensure all diameter combinations satisfy the ordering constraint
                            if not (d_b1 >= d_b2 >= d_b3 >= d_b4):
                                continue  # Skip combinations that do not meet the ordering condition
                            # Apply diameter difference condition across all combinations
                            # Ensure that no two bars exceed `max_diameter_diff`
                            diameters = [d_b1, d_b2, d_b3, d_b4]
                            diameters = [d for d in diameters if d is not None]  #Filter out None values for unused bars
                            
                            # Apply the diameter difference check across all bars
                            if not self._check_diameter_differences(diameters):
                                continue  # Skip this combination if any diameter pair exceeds max_diameter_diff`

                            n1 = 2 # This is a fixed value for every beam 
                            
                            # Iterate over possible numbers of bars in each group
                            for n2 in range(0, self.max_bars_per_layer + 1):  # n2 can be 0 or more
                                # Check spacing for the first set of bars 
                                self._check_spacing(n1, n2, d_b1, d_b2, effective_width)

                                if n1 + n2 > self.max_bars_per_layer:
                                    continue  # Skip if the total bars in layer 1 exceed the limit

                                # Calculate area for layer 1
                                A_s_layer_1 = (
                                    n1 * self.rebar_areas[d_b1] +
                                    (n2 * self.rebar_areas[d_b2] if n2 > 0 else 0*cm**2)
                                )

                                # Condition 6 and 7: Check clear spacing in layer 1
                                if n2 > 0 and not self._check_spacing(n1, n2, d_b1, d_b2, effective_width):
                                    continue

                                A_s_max = max(10*A_s_req, n1 * self.rebar_areas[self.min_long_rebar])
                                # Check if total area from layer 1 is enough for required A_s
                                # And also less than A_s_max
                                if A_s_layer_1 >= A_s_req and A_s_layer_1 <= A_s_max:
                                # if A_s_layer_1 >= A_s_req:
                                    total_as = A_s_layer_1  # Only consider layer 1
                                    total_bars = n1 + n2    # Total bars only in layer 1
                                    valid_combinations.append({
                                        'n_1': n1,
                                        'd_b1': d_b1,
                                        'n_2': n2,
                                        'd_b2': d_b2 if n2 > 0 else None,  # Display as None if n2 is 0
                                        'n_3': 0,  # No bars in layer 2
                                        'd_b3': None,
                                        'n_4': 0,  # No bars in layer 2
                                        'd_b4': None,
                                        'total_as': total_as.to('cm**2'),
                                        'total_bars': total_bars,
                                        'clear_spacing': self._clear_spacing.to('mm')
                                    })
                                else:
                                    # Track the combination with the maximum possible area (fallback)
                                    if A_s_layer_1 > max_fallback_area:
                                        max_fallback_area = A_s_layer_1
                                        best_fallback_combination = {
                                            'n_1': n1,
                                            'd_b1': d_b1,
                                            'n_2': n2,
                                            'd_b2': d_b2 if n2 > 0 else None,
                                            'n_3': 0,
                                            'd_b3': None,
                                            'n_4': 0,
                                            'd_b4': None,
                                            'total_as': A_s_layer_1.to('cm**2'),
                                            'total_bars': n1 + n2,
                                            'clear_spacing': self._clear_spacing.to('mm')
                                        }

                                # Now check combinations where bars are added in layer 2 (n3 and n4)
                                for n3 in [0, 2]:  # n3 can be 0 or fixed at 2 if present
                                    for n4 in range(0, self.max_bars_per_layer + 1):
                                        # Ensure layer 2 bars are not more than layer 1 bars
                                        if n3 + n4 > n1 + n2:
                                            continue  # Skip if layer 2 bars exceed layer 1 bars
                                        if n3 == 0 and n4 > 0:
                                            continue  # If n3 is 0, n4 must also be 0
                                        if n3 + n4 > self.max_bars_per_layer:
                                            continue  # Skip if the total bars in layer 2 exceed the limit

                                        # Layer 2 area calculation handling zero values
                                        A_s_layer_2 = (
                                            n3 * self.rebar_areas[d_b3] +
                                            (n4 * self.rebar_areas[d_b4] if n4 > 0 else 0 * cm ** 2)
                                        )
                                        # Condition 4: Area of layer 1 must be >= area of layer 2
                                        if A_s_layer_1 < A_s_layer_2:
                                            continue

                                        # Condition 6 and 7: Check clear spacing in layer 2
                                        if n4 > 0 and not self._check_spacing(n3, n4, d_b3, d_b4, effective_width):
                                            continue

                                        # Check if total area is enough for required A_s
                                        total_as = A_s_layer_1 + A_s_layer_2
                                        if total_as >= A_s_req and total_as <= A_s_max:
                                        # if total_as >= A_s_req: 
                                            total_bars = n1 + n2 + n3 + n4  # Count the total number of bars
                                            valid_combinations.append({
                                                'n_1': n1,
                                                'd_b1': d_b1,
                                                'n_2': n2,
                                                'd_b2': d_b2 if n2 > 0 else None,
                                                'n_3': n3,
                                                'd_b3': d_b3 if n3 > 0 else None,
                                                'n_4': n4,
                                                'd_b4': d_b4 if n4 > 0 else None,
                                                'total_as': total_as.to('cm**2'),
                                                'total_bars': total_bars,
                                                'clear_spacing': self._clear_spacing.to('mm')
                                            })
                                        else:
                                            # Track the combination with the maximum possible area (fallback)
                                            if total_as > max_fallback_area:
                                                max_fallback_area = total_as
                                                best_fallback_combination = {
                                                    'n_1': n1,
                                                    'd_b1': d_b1,
                                                    'n_2': n2,
                                                    'd_b2': d_b2 if n2 > 0 else None,
                                                    'n_3': n3,
                                                    'd_b3': d_b3 if n3 > 0 else None,
                                                    'n_4': n4,
                                                    'd_b4': d_b4 if n4 > 0 else None,
                                                    'total_as': total_as.to('cm**2'),
                                                    'total_bars': n1 + n2 + n3 + n4,
                                                    'clear_spacing': self._clear_spacing.to('mm')
                                                }

        # Convert valid combinations to DataFrame
        df = pd.DataFrame(valid_combinations)
        # Drop duplicate rows based on the specified columns
        df = df.drop_duplicates(subset=['n_1', 'd_b1', 'n_2', 'd_b2', 'n_3', 'd_b3', 'n_4', 'd_b4'])
        # If no valid combinations satisfy A_s_req, use the best fallback combination
        if df.empty and best_fallback_combination is not None:
            df = pd.DataFrame([best_fallback_combination])

        modified_df: pd.DataFrame = self._calculate_penalties_long_rebar(df)
        # Sort by 'Functional' to sort by the best options
        modified_df.sort_values(by=['functional'], inplace=True)
        # modified_df.drop(columns=['area_penalty','bars_penalty', 'diameter_penalty',
        #                           'layer_penalty', 'combined_penalty', 'functional'], inplace=True)
        modified_df.reset_index(drop=True, inplace=True)
        self._long_combos_df = modified_df

        return modified_df.head(10)

    def longitudinal_rebar_EN_1992_2004(self, A_s_req: PlainQuantity) -> None:
        pass
  
    def _check_spacing(self, n1:int, n2:int, d_b1: PlainQuantity, d_b2:PlainQuantity,
                       effective_width: PlainQuantity) -> bool:
        """
        Checks the clear spacing between rebars in a layer.

        Parameters:
            n1 (int): Number of bars in the first group of the layer.
            n2 (int): Number of bars in the second group of the layer.
            d_b1 (PlainQuantity): Diameter of bars in the first group of the layer.
            d_b2 (PlainQuantity): Diameter of bars in the second group of the layer.
            effective_width (PlainQuantity): The effective width available for bar placement.

        Returns:
            bool: True if the clear spacing satisfies the design limits, False otherwise.
        """
        def layer_clear_spacing(n_a: int, d_a: PlainQuantity, n_b: int, d_b: PlainQuantity) -> PlainQuantity:
            """
            Helper function to calculate clear spacing for a given layer.

            Parameters:
                n_a (int): Number of bars in the first group of the layer.
                d_a (PlainQuantity): Diameter of bars in the first group of the layer.
                n_b (int): Number of bars in the second group of the layer.
                d_b (PlainQuantity): Diameter of bars in the second group of the layer.

            Returns:
                PlainQuantity: Clear spacing for the given layer.
            """
            total_bars = n_a + n_b
            if total_bars <= 1:
                return effective_width - max(d_a, d_b)  # Clear space for one bar
            total_bar_width = n_a * d_a + n_b * d_b
            return (effective_width - total_bar_width) / (total_bars - 1)

        # Calculate the clear spacing for the layer
        self._clear_spacing = layer_clear_spacing(n1, d_b1, n2, d_b2)

        # Determine the maximum clear spacing limit
        max_clear_spacing = max(self.min_clear_spacing, self.vibrator_size, max(d_b1, d_b2))

        # Check if the clear spacing is within limits
        return self._clear_spacing >= max_clear_spacing

    def _check_diameter_differences(self, diameters: list) -> bool:
        """
        Checks that all diameter differences between bars in the list
        do not exceed max_diameter_diff.
        """
        for i, d1 in enumerate(diameters):
            for d2 in diameters[i + 1:]:
                if abs(d1 - d2) > self.max_diameter_diff:
                    return False
        return True
    
    def _calculate_penalties_long_rebar(self, df: pd.DataFrame,
                                        alpha: float =  3.5,
                                        beta: float =   0.3,
                                        gamma: float =  0.25,
                                        delta: float =  1) -> pd.DataFrame:
        """
        Calculate penalties for rebar configurations and add them as columns to the DataFrame.

        Args:
            df (pd.DataFrame): The input DataFrame containing rebar configurations.
            alpha (float): Weight for the number of bars penalty.
            beta (float): Weight for the diameter difference penalty.
            gamma (float): Weight for the layer penalty.

        Returns:
            pd.DataFrame: The modified DataFrame with penalty columns and the final functional.
        """

        # Calculate minimum bars and minimum area of steel
        min_bars = df['total_bars'].min()
        min_as = df['total_as'].min()

        # Diameter difference penalty function
        def diameter_difference_penalty(row: pd.Series) -> float:
            """
            Calculate the penalty for variation in bar diameters.
            
            Args:
                row (pd.Series): A row of the DataFrame.
                
            Returns:
                float: The standard deviation of the diameters (penalty).
            """
            diameters = []
            if row['n_1'] > 0:
                diameters.extend([row['d_b1'].magnitude] * row['n_1'])
            if row['n_2'] > 0:
                diameters.extend([row['d_b2'].magnitude] * row['n_2'])
            if row['n_3'] > 0:
                diameters.extend([row['d_b3'].magnitude] * row['n_3'])
            if row['n_4'] > 0:
                diameters.extend([row['d_b4'].magnitude] * row['n_4'])
            return np.std(diameters) if diameters else 0.0

        # Layer penalty function
        def layer_penalty(row: pd.Series) -> int:
            """
            Calculate the penalty for using additional layers of reinforcement.
            
            Args:
                row (pd.Series): A row of the DataFrame.
                
            Returns:
                int: 1 if additional layers are used, 0 otherwise.
            """
            # Check if any bars are in n_3, n_4, or future layers
            if row['n_3'] > 0 or row['n_4'] > 0:
                return 1  # Penalize if additional layers are used
            return 0  # No penalty if only the first layer is used

        
        # Calculate penalties and add them as columns
        df['area_penalty'] = df.apply(lambda row: (alpha*row['total_as']/min_as), axis=1)
        df['bars_penalty'] = beta * (df['total_bars'] - min_bars)/min_bars
        df['diameter_penalty'] = gamma * df.apply(diameter_difference_penalty, axis=1)
        df['layer_penalty'] = delta * df.apply(layer_penalty, axis=1)

        # Calculate the final functional
        df['functional'] = df.apply(
            lambda row: (
                row['area_penalty'] +
                row['bars_penalty'] +
                row['diameter_penalty'] +
                row['layer_penalty']
            ),
            axis=1
        )

        return df

    def longitudinal_rebar(self, A_s_req: PlainQuantity) -> Dict[str, Any]:
        """
        Selects the appropriate longitudinal rebar method based on the design code.
        """
        if self.beam.concrete.design_code=="ACI 318-19" or self.beam.concrete.design_code=="CIRSOC 201-25":
            return self.longitudinal_rebar_ACI_318_19(A_s_req)
        # elif self.beam.concrete.design_code=="EN 1992":
        #     return self.beam_longitudinal_rebar_EN_1992(A_s_req)
        # elif self.beam.concrete.design_code=="EHE-08":
        #     return self.beam_longitudinal_rebar_EHE_08(A_s_req)
        else:
            raise ValueError(f"Longitudinal design method not implemented \
                             for concrete type: {type(self.beam.concrete).__name__}")
    
    # Factory method to select the longitudinal rebar method
    # def transverse_rebar(self, A_v_req: PlainQuantity, V_s_req: PlainQuantity) -> DataFrame:
    #     """
    #     Selects the appropriate transverse rebar method based on the design code.
    #     """
    #     if self.beam.concrete.design_code=="ACI 318-19":
    #         return self.transverse_rebar_ACI_318_19(A_v_req, V_s_req)
    #     # elif self.beam.concrete.design_code=="EN 1992":
    #     #     return self.beam_transverse_rebar_EN_1992(A_v_req, V_s_req)
    #     # elif self.beam.concrete.design_code=="EHE-08":
    #     #     return self.beam_transverse_rebar_EHE_08(A_v_req, V_s_req)
    #     else:
    #         raise ValueError(f"Shear design method not implemented \
    #                          for concrete type: {type(self.beam.concrete).__name__}")
        



import pandas as pd
from typing import Optional, List, Any
from tabulate import tabulate
from docx import Document
from docx.shared import Pt, Cm, RGBColor
import seaborn as sns
import matplotlib.pyplot as plt
from docx.oxml import parse_xml
from docx.oxml.ns import nsdecls

CUSTOM_COLORS = {
        'blue': '#1f77b4',  # Default Matplotlib blue
        'red': '#d62728',   # Default Matplotlib red
        'dark_gray': '#323232',
        'light_gray': '#e3e3e3',
        'dark_blue':  '#073165'
    }

# Default Matplotlib and Seaborn settings
def configure_plot_settings() -> None:
    """
    Configures global settings for Matplotlib and Seaborn plots.
    """
    # Basic style for the plot
    sns.set_theme(style="whitegrid")
    sns.set_context("paper", rc={"lines.linewidth": 1.8})
    sns.set_style(rc={'axes.facecolor': '#F8F8F8'})

    # Matplotlib specific settings
    plt.rcParams.update({
        'text.usetex': True,
        'font.family': 'serif',
        'font.serif': ['Lato'],
        'axes.titlesize': 12,
        'axes.labelsize': 12,
        'xtick.labelsize': 12,
        'ytick.labelsize': 12,
        'legend.fontsize': 12,
    })

class Formatter:
    """
    Class to stlye the results in a Markdown display and a DataFrame output.
    """
    def __init__(self) -> None:
        # Define colors in the constructor
        self.green = '#439b00'
        self.red = '#d43e36'
        self.yellow = '#efc200'
        self.mid_value=0.95
        self.max_value=1
    
    def DCR(self, DCR: float) -> str:
        # Determine color based on DCR value
        if self.mid_value > DCR:
            color = self.green
        elif self.mid_value <= DCR <= self.max_value:
            color = self.yellow
        else:
            color = self.red
        
        return f"$\\color{{{color}}}{{\\text{{DCR}}={round(DCR,2)}}}$"
    
    def DCR_value(self, DCR: float) -> str:
        # Determine color based on DCR value
        if self.mid_value > DCR:
            color = self.green
        elif self.mid_value <= DCR <= 1:
            color = self.yellow
        else:
            color = self.red
        
        return f"$\\color{{{color}}}{{{round(DCR,2)}}}$"
    
    def is_lower(self, value1: float, value2: float) -> str:
        # Compare two values and return the appropriate formatted output
        if value1 < value2:
            return r"$\color{" + self.green + r"}{\, \checkmark}$"  # Green checkmark
        else:
            return r"$\color{" + self.red + r"}{\, \times}$"  # Red cross
    
    def is_greater(self, value1: float, value2: float) -> str:
        # Compare two values and return the appropriate formatted output
        if value1 > value2:
            return r"$\color{" + self.green + r"}{\, \checkmark}$"  # Green checkmark
        else:
            return r"$\color{" + self.red + r"}{\, \times}$"  # Red cross
    # Formatting functions outside the class

    def DCR_value_df(self,DCR: Optional[float]) -> str:
        # If DCR is None or not a number, return an empty string
        if DCR is None or not isinstance(DCR, (int, float)):
            return ""

        # Determine the color based on DCR value
        if self.mid_value > DCR:
            return f"color: {self.green}"
        elif self.mid_value <= DCR <= self.max_value:
            return f"color: {self.yellow}"
        else:
            return f"color: {self.red}"

    def apply_DCR_style(self,value: float) -> str:
        formatted_value = round(value, 2) if isinstance(value, (int, float)) else value
        return self.DCR_value_df(formatted_value)

    def color_DCR_df(self, df: pd.DataFrame, DCR_columns: list) -> pd.io.formats:
        """
        Apply color styling to specified DCR-related columns in the DataFrame.
        
        :param df: DataFrame with DCR values to color.
        :param DCR_columns: List of column names to apply the DCR styling to.
        :return: A styled DataFrame with colored DCR values.
        """
        return df.style.map(self.apply_DCR_style, subset=DCR_columns).format(precision=2)

class TablePrinter:
    """
    A class for printing tables with customizable formatting options.
    
    Attributes:
    -----------
    title : Optional[str]
        Optional title displayed above the printed table.

    Methods:
    --------
    print_table_data(data: List[List[Any]], headers: List[str], tablefmt: str = "fancygrid", numalign: str = "right")
      -> None
        Prints table data with customizable formatting.
        
    print_table_min_max(data: List[List[Any]], headers: List[str], tablefmt: str = "fancygrid", numalign: str = "right")
      -> None
        Prints table data with column alignment for minimum and maximum values.
    """
    
    def __init__(self, title: Optional[str] = None) -> None:
        """
        Initializes the TablePrinter with an optional title.
        
        Parameters:
        -----------
        title : Optional[str], default=None
            Title to be displayed above the table, if provided.
        """
        self.title = title

    def print_table_data(self, data: dict[str, list[Any]], headers: str, tablefmt: str = "fancygrid",
                         numalign: str = "right") -> None:
        """
        Prints table data with customizable formatting options.
        
        Parameters:
        -----------
        data : List[List[Any]]
            The data to be printed in table format, where each inner list represents a row.
        
        headers : List[str]
            Column headers for the table.
        
        tablefmt : str, default="fancygrid"
            Table formatting style supported by tabulate (e.g., "plain", "grid", "fancygrid").
        
        numalign : str, default="right"
            Number alignment in the table. Common options are "right", "center", or "left".
        
        Returns:
        --------
        table:  Return the formatted table string
        """
        # if self.title:
        #     print(f"======= {self.title} ======= \n")
        
        colalign = ("left", "center", "right", "left")
        table = tabulate(
            data,
            headers=headers,
            tablefmt=tablefmt,
            numalign=numalign,
            colalign=colalign
        )
        print(table, '\n')
        return None  
        
    def print_table_min_max(self, data: dict[str, list[Any]], headers: str, tablefmt: str = "fancygrid",
                            numalign: str = "right") -> None:
        """
        Prints table data with column alignment for minimum and maximum values.
        
        Parameters:
        -----------
        data : List[List[Any]]
            The data to be printed in table format, where each inner list represents a row.
        
        headers : List[str]
            Column headers for the table.
        
        tablefmt : str, default="fancygrid"
            Table formatting style supported by tabulate (e.g., "plain", "grid", "fancygrid").
        
        numalign : str, default="right"
            Number alignment in the table. Common options are "right", "center", or "left".
        
        Returns:
        --------
        None
        """
        if self.title:
            print(f"======= {self.title} ======= \n")
        
        colalign = ("left", "center", "center", "center", "center", "center", "left")
        table = tabulate(
            data,
            headers=headers,
            tablefmt=tablefmt,
            numalign=numalign,
            colalign=colalign
        )
        print(table, '\n')

class DocumentBuilder:
    """
    A class to build and style a Word document, including adding headings and tables from data frames.
    
    Attributes:
    -----------
    title : str
        Title of the document, displayed at the beginning.
        
    font_name : str
        Default font name for the document text.
        
    font_size : int
        Default font size for the document text.
        
    doc : Document
        The Document object representing the Word document being built.

    Methods:
    --------
    set_document_style() -> None
        Configures the document's default style with the specified font name and size.
        
    add_heading(text: str, level: int) -> None
        Adds a heading to the document at the specified level.
        
    set_col_widths(table: 'docx.table.Table', column_widths: List[Cm]) -> None
        Sets the width for each column in a given table.
        
    add_table(df: pd.DataFrame, column_widths: List[Cm]) -> None
        Adds a table to the document, with data from a DataFrame and custom column widths.
        
    save(filename: str) -> None
        Saves the document to a specified filename.
    """
    
    def __init__(self, title: str, font_name: str = 'Lato', font_size: int = 9) -> None:
        """
        Initializes the DocumentBuilder with a title, font name, and font size.
        
        Parameters:
        -----------
        title : str
            Title to be displayed in the document.
            
        font_name : str, default='Lato'
            Font name to be used for the document text.
            
        font_size : int, default=9
            Font size for the document text.
        """
        self.doc = Document()
        self.title = title
        self.font_name = font_name
        self.font_size = font_size
        self.set_document_style()
        self.set_page_size()

    def set_document_style(self) -> None:
        """
        Sets the default style of the document, applying the font name and size.
        
        Returns:
        --------
        None
        """
        style = self.doc.styles['Normal']
        font = style.font
        font.name = self.font_name
        font.size = Pt(self.font_size)

    def set_page_size(self) -> None:
        """
        Sets the page size to A4 by default.
        
        Returns:
        --------
        None
        """
        section = self.doc.sections[0]
        section.page_height = Cm(29.7)
        section.page_width = Cm(21.0)

    def add_heading(self, text: str, level: int) -> None:
        """
        Adds a heading to the document at the specified level.
        
        Parameters:
        -----------
        text : str
            The text for the heading.
            
        level : int
            The heading level (e.g., 0 for title, 1 for main headings, etc.).
        
        Returns:
        --------
        None
        """
        heading = self.doc.add_heading(text, level=level)
        heading.paragraph_format.space_before = Pt(0)  # Remove space after the paragraph

    def add_text(self, text: str) -> None:
        """ Adds a paragraph to the document """
        self.doc.add_paragraph(text)
        
    def set_col_widths(self, table: Any, column_widths: List[Cm]) -> None:
        """
        Sets the width of each column in the table.
        
        Parameters:
        -----------
        table : docx.table.Table
            The table object whose column widths need to be set.
            
        column_widths : List[Cm]
            List of widths for each column, in centimeters.
        
        Returns:
        --------
        None
        """
        for row in table.rows:
            for idx, width in enumerate(column_widths):
                row.cells[idx].width = width

    def add_table(self, df: pd.DataFrame, column_widths: List[Cm]) -> None:
        """
        Adds a table to the document, populated with data from a DataFrame.
        
        Parameters:
        -----------
        df : pd.DataFrame
            The DataFrame containing the data for the table.
            
        column_widths : List[Cm]
            List of column widths for the table.
        
        Returns:
        --------
        Table
        """
        # Initialize the table with the DataFrame dimensions
        table = self.doc.add_table(rows=df.shape[0] + 1, cols=df.shape[1])
        # Apply table style
        table.style = "Light Shading"
        # Set column widths
        self.set_col_widths(table, column_widths)
        # Populate column headers
        for j in range(df.shape[1]):
            table.cell(0, j).text = str(df.columns[j])
        # Populate table data
        for i in range(df.shape[0]):
            for j in range(df.shape[1]):
                cell = df.iat[i, j]
                table.cell(i + 1, j).text = str(cell)
        # Customize style for first column
        for row in table.rows[1:]:
            cell = row.cells[0]
            run = cell.paragraphs[0].runs[0]
            run.font.bold = False
        # Add a spacer paragraph after the table
        spacer = self.doc.add_paragraph()
        spacer.paragraph_format.space_after = Pt(0)  # Remove space after the paragraph

        return None

    def add_table_data(self, df: pd.DataFrame) -> None:
        column_widths = [Cm(12), Cm(2), Cm(2), Cm(2)]
        self.add_table(df, column_widths)

    def add_table_dcr(self, df: pd.DataFrame) -> None:
        """
        Adds a table using `add_table` and modifies the last row's DCR cell color.

        - Green if DCR < 1
        - Red if DCR >= 1
        """
        column_widths = [Cm(12), Cm(2), Cm(2), Cm(2)]
        self.add_table(df, column_widths)
        # Get the last table added to the document
        table = self.doc.tables[-1]  # Retrieve the most recent table

        # Apply color to DCR cell (third column of the last row)
        last_row_idx = df.shape[0]  # Last row index
        dcr_column_idx = 2  # Third column index (zero-based)

        dcr_value = float(df.iat[-1, dcr_column_idx])  # Extract DCR value
        if dcr_value < 1:
            shading_color = "C6EFCE"  # Green
            font_color = "006100"  # Green font
        else:
            shading_color = "FFC7CE"  # Red
            font_color = "9C0006"  # Red font

        # Apply shading and font color to the entire last row
        for cell in table.rows[last_row_idx].cells:
            # Apply background color
            cell._element.get_or_add_tcPr().append(
                parse_xml(f'<w:shd {nsdecls("w")} w:fill="{shading_color}"/>')
            )
            # Apply font color
            for paragraph in cell.paragraphs:
                for run in paragraph.runs:
                    run.font.color.rgb = RGBColor.from_string(font_color)
    
    def add_table_min_max(self, df: pd.DataFrame) -> None:
        column_widths = [Cm(10), Cm(2), Cm(2), Cm(2), Cm(2), Cm(1)]
        self.add_table(df, column_widths)


    def save(self, filename: str) -> None:
        """
        Saves the document to a specified file.
        
        Parameters:
        -----------
        filename : str
            The filename (including path) where the document will be saved.
        
        Returns:
        --------
        None
        """
        self.doc.save(filename)



    #  Examples to run in a Jupyter Notebook
    # formatter = Formatter()
    # display(Markdown(formatter.DCR(0.85)))
    # display(Markdown(formatter.DCR_value(0.85)))
    # display(Markdown(formatter.is_lower(0.85,1)))
    # display(Markdown(formatter.is_greater(0.85,1)))



from dataclasses import dataclass, field
from typing import Optional, TYPE_CHECKING, Dict, Any, List, Optional

from mento.settings import Settings
from mento.material import Concrete, SteelBar
from mento.forces import Forces

if TYPE_CHECKING:
    from pint.facets.plain import PlainQuantity
    from mento.node import Node
    
@dataclass
class Section:
    _id: int
    settings: Settings = field(default_factory=Settings)
    _last_id: int = field(default=0, init=False, repr=False)  # Class variable to keep track of last assigned ID
    node: Optional["Node"] = None  # Use forward declaration with Optional["Node"]
    label: Optional[str] = None

    def __init__(self, label: Optional[str], concrete: Concrete, steel_bar: SteelBar, settings: Optional[Dict[str, Any]] = None) -> None:
        # Initialize the private ID automatically
        Section._last_id += 1  # Increment the class variable for the next ID
        self._id = Section._last_id  # Assign the next available ID
        self.concrete = concrete
        self.steel_bar = steel_bar
        self.label = label
        
        # Initialize with default settings
        self.settings = Settings(concrete) 
        if settings:
            self.settings.update(settings)  # Update with any provided settings
        self.c_c: PlainQuantity = self.settings.get_setting('clear_cover')
        self._stirrup_d_b: PlainQuantity = self.settings.get_setting('stirrup_diameter_ini')
        self._layers_spacing: PlainQuantity = self.settings.get_setting('layers_spacing')
        

    @property
    def id(self) -> int:
        """Read-only property to access the private _id."""
        return self._id

    @property
    def get_settings(self) -> Dict[str, Any]:
        """Returns the complete current settings as a dictionary."""
        return self.settings.settings  # Access the settings dictionary from the Settings instance
       
    def update_settings(self, new_settings: Dict[str, Any]) -> None:
        """Updates settings with new values."""
        self.settings.update(new_settings)

    def check_shear(self, forces: List[Forces]) -> None:
        pass

    def design_shear(self, forces: List[Forces]) -> None:
        pass

    def check_flexure(self, forces: List[Forces]) -> None:
        pass

    def design_flexure(self, forces: List[Forces]) -> None:
        pass

    def shear_results_detailed(self, force: Optional[Forces] = None) -> None:
        pass

    def shear_results_detailed_doc(self, force: Optional[Forces] = None) -> None:
        pass

    def flexure_results_detailed(self, force: Optional[Forces] = None) -> None:
        pass

    def flexure_results_detailed_doc(self, force: Optional[Forces] = None) -> None:
        pass
    
    # Beam results for Jupyter Notebook
    @property
    def results(self) -> None:
        return self.results

